(*****************************************************************************
 *          HOL-OCL - a shallow embedding of OCL in Isabelle/HOL              
 *                                                                            
 * encoding_example.thy - 
 * Copyright (C) 2006 Achim D. Brucker <brucker@inf.ethz.ch>   
 *                                                                            
 * This File is part of HOL-OCL.                                              
 *                                                                            
 * HOL-OCL is free software; you can redistribute it and/or modify it under   
 * the terms of the GNU General Public License as published by the Free       
 * Software Foundation; either version 2 of the License, or (at your option)  
 * any later version.                                                         
 *                                                                            
 * HOL-OCL is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS 
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more 
 * details.                                                              
 *                                                                            
 * You should have received a copy of the GNU General Public License along    
 * with this program; if not, write to the Free Software Foundation, Inc.,    
 * 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.                  
 ******************************************************************************)


header {* Encoding \UML/\OCL into \HOL by Example *}

(*<*)
theory encoding_example  
imports 
  OCL
begin
(*>*)

text_raw{* \label{thy:encoding_example} *} 

chapter {* A Simple Example *}
text {* 
In this section we present the details of our encoding by an tiny
example. We use the class diagram presented in \autoref{fig:encoding}
together with the \OCL specification shown in
\autoref{lst:encoding}. Our Example consists only out of two classes
with an inheritance relationship. Class \inlineocl{A} has one
attribute \inlineocl{i} of type \inlineocl{Integer} that is constraint
by an invariant. Further, class \inlineocl{A} has a method
\inlineocl{f(x:Integer):Integer} which is described by a pre-condition
and post-condition. Class \inlineocl{B} only has one attribute
\inlineocl{j} of type \inlineocl{B}. Note that this is, in contrast to
the attribute \inlineocl{i} of class \inlineocl{A}, an object type.
In the following we describe which definition and theorems are
generated during the import of this simple class diagram into \holocl.

\begin{figure}
  \begin{center}
    \includegraphics[scale=.6]{figures/encoding_example}
  \end{center}
  \caption{A simple Encoding Example: Data Model\label{fig:encoding}}
\end{figure}

\lstinputlisting[style=displayocl,float,caption={A simple Encoding Example: \OCL specification},
                 label={lst:encoding}]{encoding_example.ocl}
*}

chapter {* Importing \acs{uml}/\acs{ocl} models *} 

text {*
 \acs{uml}\acs{ocl} models can be directly imported into \holocl if
  they are available in \acs{xmi} format, e.g.,
*}
load_xmi "encoding_example_ocl.xmi"
text {* 
\noindent imports the \acs{uml}/\acs{ocl} specification of our
example. During the the import, different definitions and theorems are
generated into the current theory context. \holocl{} organizes these
automatically generated stuff into a hierarchy of namespaces, i.e., in
our example:
\begin{description}
\item[@{text UML_OCL}:] This is the \emph{standard} namespace for internal
  definitions, i.e., definitions that normally should not be needed by
  the users. This namespace is divided as follows:
  \begin{description}
  \item[@{text univ.model}:] contains the basic type definitions of classes and the type of the 
    universe
  \item[@{text level0.model}:] contains the level 0 definitions of the model. 
  \item[@{text level1.model}:] contains the level 1 definitions of the model.
  \end{description}
\item[@{text model}:] It contains all definitions a user would expect, namely all the 
  level 2 stuff which ties the \UML and \OCL part together.
\end{description}
In all namespaces, ``@{text model}'' is replaced by the top-level
\acs{uml} package of the encoded model, i.e., for our example 
``@{text encoding}''. Thus, in our example all level 0 definitions are defined
within ``@{text UML_OCL.level0.encoding}''.
  *}

text_raw {*
  \let\oldisanewline\isanewline
  \let\oldisa\isa  
  \renewcommand*{\isanewline}{\notag\\}
  \renewcommand*{\isa}[1]{#1}
 *}

section {* Encoding Level 0: *}
subsection {* The Universe Type *}
 text {* 

   The first major step of encoding classes into the base \OCL
   universe is to construct the universe type, but first we have to
   construct  the core types for classes. For each class we construct
   a ``identifier'' type which is crucial for our strong type 
   discipline:
   \begin{align} 
    @{typ [long_names] "UML_OCL.univ.encoding.A_key"} 
          &= @{typeof [] "(obj::UML_OCL.univ.encoding.A_key)"} \notag
    \intertext{the base type of the class, describing its locally defined
    attributes}
    @{typ [long_names] "UML_OCL.univ.encoding.A_base_type"} 
          &= @{typeof [] "(obj::UML_OCL.univ.encoding.A_base_type)"} \notag
    \intertext{the type of the class, describing the inherited type together with the 
      base type}
    @{typ [long_names] "UML_OCL.univ.encoding.A_type"}  
          &= @{typeof [] "(obj::UML_OCL.univ.encoding.A_type)"} \notag\\
    @{typ [long_names] "'a UML_OCL.univ.encoding.A_ext"} 
          &= @{typeof [] "(obj::('a UML_OCL.univ.encoding.A_ext))"} \notag
  \end{align}
  These types are analogous defined for the inherited class  \inlineocl{B}:
   \renewcommand*{\isanewline}{\notag\\&\quad}
  \begin{align} 
    @{typ [long_names] "UML_OCL.univ.encoding.B_key"}       
          &= @{typeof [] "(obj::UML_OCL.univ.encoding.B_key)"} \notag\\
    @{typ [long_names] "UML_OCL.univ.encoding.B_base_type"}      
          &= @{typeof [] "(obj::UML_OCL.univ.encoding.B_base_type)"} \notag\\
    @{typ [long_names] "UML_OCL.univ.encoding.B_type"}      
          &= @{typeof [] "(obj::UML_OCL.univ.encoding.B_type)"} \notag\\
    @{typ [long_names] "'a UML_OCL.univ.encoding.B_ext"}    
          &= @{typeof [break,margin=65] "(obj::('a UML_OCL.univ.encoding.B_ext))" }\notag
  \end{align}
  \renewcommand*{\isanewline}{\notag\\}
  Together with the pre-defined \acs{ocl} universe type they build the
  basis for the model specific universe type:
  \begin{multline}\notag
    @{typ [long_names] "('a,'b,'c) U_encoding_example"} \\
    = 
    @{typeof [] "(obj::(('a,'b,'c) U_encoding_example))"}
  \end{multline}
  Now we can encode types for classes represented in the universe, e.g. for class \inlineocl{A}:
  \begin{multline}\notag 
    @{typ [long_names] "('a,'b) UML_OCL.univ.encoding.A"} \\
     =  @{typeof [] "(obj::(('a,'b) UML_OCL.univ.encoding.A))"} 
   \end{multline}
   and for class  \inlineocl{B}:
   \begin{multline}\notag 
    @{typ [long_names] "('a) UML_OCL.univ.encoding.B"} \\
    = @{typeof [] "(obj::(('a) UML_OCL.univ.encoding.B))"}
    \end{multline}
*} 

subsection {* Defining the Basic Methods over Classes *}
 text {* 
For each class we define two type tests for objects, one for testing 
if a object is of a specific type or suptype (i.e., this test is true, if the 
object is of the given kind):
	   \begin{multline}
          @{const [long_names] UML_OCL.level0.encoding.is_A}  \isasymColon      \notag\\
	   @{typeof [break,margin=80] "UML_OCL.level0.encoding.is_A"}        \notag \shoveleft{}\\ 
	    \shoveright{} @{thm [break,margin=80] UML_OCL.level0.encoding.is_A_def} \\
         \end{multline}
and one for testing if a object is exactly the requested type 
         \begin{multline}
	  @{const [long_names] UML_OCL.level0.encoding.isType_A}  \isasymColon   \notag\\
	   @{typeof [break,margin=80] "UML_OCL.level0.encoding.isType_A"}    \notag \shoveleft{}\\ 
	    \shoveright{} @{thm [break,margin=80] UML_OCL.level0.encoding.isType_A_def}\\
         \end{multline}
Further, we define a type for universe instances:
         \begin{multline}
	  @{const [long_names] UML_OCL.level0.encoding.is_A_univ}  \isasymColon  \notag\\
	   @{typeof [break,margin=80] "UML_OCL.level0.encoding.is_A_univ"}   \notag \shoveleft{}\\ 
	    \shoveright{} @{thm [break,margin=80] UML_OCL.level0.encoding.is_A_univ_def}\\
         \end{multline}
	 Analogous we define for the inherited class  \inlineocl{B}:
	   \begin{multline}
          @{const [long_names] UML_OCL.level0.encoding.is_B}  \isasymColon      \notag\\
	   @{typeof [break,margin=80] "UML_OCL.level0.encoding.is_B"}        \notag \shoveleft{}\\ 
	    \shoveright{} @{thm [break,margin=80] UML_OCL.level0.encoding.is_B_def} \\
         \end{multline}
         \begin{multline}
	  @{const [long_names] UML_OCL.level0.encoding.isType_B}  \isasymColon   \notag\\
	   @{typeof [break,margin=80] "UML_OCL.level0.encoding.isType_B"}    \notag \shoveleft{}\\ 
	    \shoveright{} @{thm [break,margin=80] UML_OCL.level0.encoding.isType_B_def}\\
         \end{multline}
         \begin{multline}
	  @{const [long_names] UML_OCL.level0.encoding.is_B_univ}  \isasymColon  \notag\\
	   @{typeof [break,margin=80] "UML_OCL.level0.encoding.is_B_univ"}   \notag \shoveleft{}\\ 
	    \shoveright{} @{thm [break,margin=80] UML_OCL.level0.encoding.is_B_univ_def}\\
         \end{multline}
*}
text {*
       We define a functions for embeding a class into the corresponding universe
          \begin{multline}
            @{const [long_names] UML_OCL.level0.encoding.mk_A}  \isasymColon       \notag\\
	    @{typeof [break,margin=80] "UML_OCL.level0.encoding.mk_A"}  \notag \shoveleft{}\\ 
	    \shoveright{} @{thm [break,margin=80] UML_OCL.level0.encoding.mk_A_def}  \\
	  \end{multline}
	  and one for extracting an object out of a given universe:
	  \begin{multline}
          @{const [long_names] UML_OCL.level0.encoding.get_A}  \isasymColon      \notag\\
	   @{typeof [break,margin=80] "UML_OCL.level0.encoding.get_A"}      \notag \shoveleft{}\\ 
	    \shoveright{} @{thm [break,margin=80] UML_OCL.level0.encoding.get_A_def}\\
	   \end{multline}
	   These two functions can also be understood as conversion between object types
	   and universe types.

 	 Analogous we define for the inherited class  \inlineocl{B}:
          \begin{multline}
            @{const [long_names] UML_OCL.level0.encoding.mk_B}  \isasymColon       \notag\\
	    @{typeof [break,margin=80] "UML_OCL.level0.encoding.mk_B"}  \notag \shoveleft{}\\ 
	    \shoveright{} @{thm [break,margin=80] UML_OCL.level0.encoding.mk_B_def}  \\
	  \end{multline}
	  \begin{multline}
          @{const [long_names] UML_OCL.level0.encoding.get_B}  \isasymColon      \notag\\
	   @{typeof [break,margin=80] "UML_OCL.level0.encoding.get_B"}      \notag \shoveleft{}\\ 
	    \shoveright{} @{thm [break,margin=80] UML_OCL.level0.encoding.get_B_def}\\
	   \end{multline}
*}
text {*
        For converting along type hierarchies (i.e. down-casting and up-casting) we
	provide for each class two functions for the conversion to the direct predecessor 
	and direct successor in the class hierarchy:
        \begin{multline}
          @{const [long_names] UML_OCL.level0.encoding.A_2_OclAny}  \isasymColon  \notag\\
	   @{typeof [break,margin=80] "UML_OCL.level0.encoding.A_2_OclAny"}   \notag \shoveleft{}\\ 
	    \shoveright{} @{thm [break,margin=80] UML_OCL.level0.encoding.A_2_OclAny_def}\\
         \end{multline}
         \begin{multline}
          @{const [long_names] UML_OCL.level0.encoding.OclAny_2_A} \isasymColon   \notag\\
	     @{typeof [break,margin=80] "UML_OCL.level0.encoding.OclAny_2_A"}  \notag \shoveleft{}\\ 
	    \shoveright{} @{thm [break,margin=80] UML_OCL.level0.encoding.OclAny_2_A_def}\\
         \end{multline}
	 Analogous we define for the inherited class  \inlineocl{B}:
         \begin{multline}
          @{const [long_names] UML_OCL.level0.encoding.B_2_A}  \isasymColon  \notag\\
	   @{typeof [break,margin=80] "UML_OCL.level0.encoding.B_2_A"}   \notag \shoveleft{}\\ 
	    \shoveright{} @{thm [break,margin=80] UML_OCL.level0.encoding.B_2_A_def}\\
         \end{multline}
         \begin{multline}
          @{const [long_names] UML_OCL.level0.encoding.A_2_B} \isasymColon   \notag\\
	     @{typeof [break,margin=80] "UML_OCL.level0.encoding.A_2_B"}  \notag \shoveleft{}\\ 
	    \shoveright{} @{thm [break,margin=80] UML_OCL.level0.encoding.A_2_B_def}\\
         \end{multline} 

*}
 text {*
   Finally, we provide accesor functions for the class attributes:
         \begin{multline}
          @{const [long_names] UML_OCL.level0.encoding.A.i} \isasymColon   \notag\\
	     @{typeof [break,margin=80] "UML_OCL.level0.encoding.A.i"}  \notag \shoveleft{}\\ 
	    \shoveright{} @{thm [break,margin=80] UML_OCL.level0.encoding.A.i_def}\\
         \end{multline} 

         \begin{multline}
          @{const [long_names] UML_OCL.level0.encoding.B.j} \isasymColon   \notag\\
	     @{typeof [break,margin=80] "UML_OCL.level0.encoding.B.j"}  \notag \shoveleft{}\\ 
	    \shoveright{} @{thm [break,margin=80] UML_OCL.level0.encoding.B.j_def}\\
         \end{multline} 
*}

subsection {* Proving Properties *} 
   text {* Already on this level, we can prove important properties of our encoding. For 
     example, 
           the following theorem  
           ``@{text UML_OCL.level0.encoding.is_A_univ_implies_OclAny_univ}'' shows, that on the
	   universe types, every object of class \inlineocl{A} is also an object of class
	   \inlineocl{OclAny}: 
	   \begin{multline}\notag
                  @{thm [break,margin=80] UML_OCL.level0.encoding.is_A_univ_implies_OclAny_univ } \\
           \end{multline}
	   Using @{text get_mk_A_id}  and @{text get_mk_A_id} we can convert ``lossless'' between our 
	   class type representation  and universe representation, this is shown by 
           ``@{text UML_OCL.level0.encoding.get_mk_A_id}'':
           \begin{multline}\notag
                  @{thm [break,margin=80] UML_OCL.level0.encoding.get_mk_A_id }  \\
           \end{multline}
           and ``@{text UML_OCL.level0.encoding.mk_get_A_id}'': 
           \begin{multline}\notag
                  @{thm [break,margin=80] UML_OCL.level0.encoding.mk_get_A_id }  \\
           \end{multline}
           and ``@{text UML_OCL.level0.encoding.is_A_univ_implies_is_get}''  shows that 
	   a successfull type test on the universe types also implies a successfull type
	   test on the class types: 
           \begin{multline}\notag
                @{thm [break,margin=80] UML_OCL.level0.encoding.is_A_univ_implies_is_get } \\
           \end{multline}\notag
	   The theorem ``@{text UML_OCL.level0.encoding.is_mk_A}'' shows that the constructor 
	   creates an object that is of 
	   type \inlineocl{A}:
           \begin{multline}\notag
                  @{thm [break,margin=80] UML_OCL.level0.encoding.is_mk_A }  \\
           \end{multline}
	   and also satisfies the test of its super-type, e.g., \inlineocl{OclAny}. This is shown 
	   by the following theorem 
	   ``@{text UML_OCL.level0.encoding.is_OclAny_mk_A}'': 
           \begin{multline}\notag
                  @{thm [break,margin=80] UML_OCL.level0.encoding.is_OclAny_mk_A}  \\
           \end{multline}
	   ``@{text UML_OCL.level0.encoding.cast_A_id}'' shows that we can successfully
	   downcast objects: 
           \begin{multline}\notag
                  @{thm [break,margin=80] UML_OCL.level0.encoding.cast_A_id } \\
           \end{multline}

           Analogous, these properties are also shown for class \inlineocl{B}: \\
           @{text UML_OCL.level0.encoding.is_B_univ_implies_A_univ}: 
	   \begin{multline}\notag
                  @{thm [break,margin=80] UML_OCL.level0.encoding.is_B_univ_implies_A_univ } \\
           \end{multline}
           @{text UML_OCL.level0.encoding.is_B_univ_implies_is_get}: 
           \begin{multline}\notag
                @{thm [break,margin=80] UML_OCL.level0.encoding.is_B_univ_implies_is_get } \\
           \end{multline}\notag
           @{text UML_OCL.level0.encoding.get_mk_B_id}:
           \begin{multline}\notag
                  @{thm [break,margin=80] UML_OCL.level0.encoding.get_mk_B_id }  \\
           \end{multline}
           @{text UML_OCL.level0.encoding.is_mk_B}: 
           \begin{multline}\notag
                  @{thm [break,margin=80] UML_OCL.level0.encoding.is_mk_B }  \\
           \end{multline}
           @{text UML_OCL.level0.encoding.is_A_mk_B}: 
           \begin{multline}\notag
                  @{thm [break,margin=80] UML_OCL.level0.encoding.is_A_mk_B}  \\
           \end{multline}
           @{text UML_OCL.level0.encoding.mk_get_B_id}: 
           \begin{multline}\notag
                  @{thm [break,margin=80] UML_OCL.level0.encoding.mk_get_B_id }  \\
           \end{multline}
           @{text UML_OCL.level0.encoding.cast_B_id}: 
           \begin{multline}\notag
                  @{thm [break,margin=80] UML_OCL.level0.encoding.cast_B_id } \\
           \end{multline}
 *}

section {* Encoding Level 1: *}

 text {* The encoding on level 1 follows strictly the schema from level 0, i.e,
we also start  two type tests for objects, one for testing 
if a object is of a specific type or subtype (i.e., this test is true, if the 
object is of the given kind):
	   \begin{multline}
          @{const [long_names] UML_OCL.level1.encoding.is_A}  \isasymColon      \notag\\
	   @{typeof [break,margin=80] "UML_OCL.level1.encoding.is_A"}        \notag \shoveleft{}\\ 
	    \shoveright{} @{thm [break,margin=80] UML_OCL.level1.encoding.is_A_def} \\
         \end{multline}
and one for testing if a object is exactly the requested type 
         \begin{multline}
	  @{const [long_names] UML_OCL.level1.encoding.isType_A}  \isasymColon   \notag\\
	   @{typeof [break,margin=80] "UML_OCL.level1.encoding.isType_A"}    \notag \shoveleft{}\\ 
	    \shoveright{} @{thm [break,margin=80] UML_OCL.level1.encoding.isType_A_def}\\
         \end{multline}
Further, we define a type for universe instances:
         \begin{multline}
	  @{const [long_names] UML_OCL.level1.encoding.is_A_univ}  \isasymColon  \notag\\
	   @{typeof [break,margin=80] "UML_OCL.level1.encoding.is_A_univ"}   \notag \shoveleft{}\\ 
	    \shoveright{} @{thm [break,margin=80] UML_OCL.level1.encoding.is_A_univ_def}\\
         \end{multline}
	 Analogous we define for the inherited class  \inlineocl{B}:
	   \begin{multline}
          @{const [long_names] UML_OCL.level1.encoding.is_B}  \isasymColon      \notag\\
	   @{typeof [break,margin=80] "UML_OCL.level1.encoding.is_B"}        \notag \shoveleft{}\\ 
	    \shoveright{} @{thm [break,margin=80] UML_OCL.level1.encoding.is_B_def} \\
         \end{multline}
         \begin{multline}
	  @{const [long_names] UML_OCL.level1.encoding.isType_B}  \isasymColon   \notag\\
	   @{typeof [break,margin=80] "UML_OCL.level1.encoding.isType_B"}    \notag \shoveleft{}\\ 
	    \shoveright{} @{thm [break,margin=80] UML_OCL.level1.encoding.isType_B_def}\\
         \end{multline}
         \begin{multline}
	  @{const [long_names] UML_OCL.level1.encoding.is_B_univ}  \isasymColon  \notag\\
	   @{typeof [break,margin=80] "UML_OCL.level1.encoding.is_B_univ"}   \notag \shoveleft{}\\ 
	    \shoveright{} @{thm [break,margin=80] UML_OCL.level1.encoding.is_B_univ_def}\\
         \end{multline}
*}
text {*
       We define a functions for embedding a class into the corresponding universe
          \begin{multline}
            @{const [long_names] UML_OCL.level1.encoding.mk_A}  \isasymColon       \notag\\
	    @{typeof [break,margin=80] "UML_OCL.level1.encoding.mk_A"}  \notag \shoveleft{}\\ 
	    \shoveright{} @{thm [break,margin=80] UML_OCL.level1.encoding.mk_A_def}  \\
	  \end{multline}
	  and one for extracting an object out of a given universe:
	  \begin{multline}
          @{const [long_names] UML_OCL.level1.encoding.get_A}  \isasymColon      \notag\\
	   @{typeof [break,margin=80] "UML_OCL.level1.encoding.get_A"}      \notag \shoveleft{}\\ 
	    \shoveright{} @{thm [break,margin=80] UML_OCL.level1.encoding.get_A_def}\\
	   \end{multline}
	   These two functions can also be understood as conversion between object types
	   and universe types.

 	 Analogous we define for the inherited class  \inlineocl{B}:
          \begin{multline}
            @{const [long_names] UML_OCL.level1.encoding.mk_B}  \isasymColon       \notag\\
	    @{typeof [break,margin=80] "UML_OCL.level1.encoding.mk_B"}  \notag \shoveleft{}\\ 
	    \shoveright{} @{thm [break,margin=80] UML_OCL.level1.encoding.mk_B_def}  \\
	  \end{multline}
	  \begin{multline}
          @{const [long_names] UML_OCL.level1.encoding.get_B}  \isasymColon      \notag\\
	   @{typeof [break,margin=80] "UML_OCL.level1.encoding.get_B"}      \notag \shoveleft{}\\ 
	    \shoveright{} @{thm [break,margin=80] UML_OCL.level1.encoding.get_B_def}\\
	   \end{multline}
*}
text {*
        For converting along type hierarchies (i.e. down-casting and up-casting) we
	provide for each class two functions for the conversion to the direct predecessor 
	and direct successor in the class hierarchy:
        \begin{multline}
          @{const [long_names] UML_OCL.level1.encoding.A_2_OclAny}  \isasymColon  \notag\\
	   @{typeof [break,margin=80] "UML_OCL.level1.encoding.A_2_OclAny"}   \notag \shoveleft{}\\ 
	    \shoveright{} @{thm [break,margin=80] UML_OCL.level1.encoding.A_2_OclAny_def}\\
         \end{multline}
         \begin{multline}
          @{const [long_names] UML_OCL.level1.encoding.OclAny_2_A} \isasymColon   \notag\\
	     @{typeof [break,margin=80] "UML_OCL.level1.encoding.OclAny_2_A"}  \notag \shoveleft{}\\ 
	    \shoveright{} @{thm [break,margin=80] UML_OCL.level1.encoding.OclAny_2_A_def}\\
         \end{multline}
	 Analogous we define for the inherited class  \inlineocl{B}:
         \begin{multline}
          @{const [long_names] UML_OCL.level1.encoding.B_2_A}  \isasymColon  \notag\\
	   @{typeof [break,margin=80] "UML_OCL.level1.encoding.B_2_A"}   \notag \shoveleft{}\\ 
	    \shoveright{} @{thm [break,margin=80] UML_OCL.level1.encoding.B_2_A_def}\\
         \end{multline}
         \begin{multline}
          @{const [long_names] UML_OCL.level1.encoding.A_2_B} \isasymColon   \notag\\
	     @{typeof [break,margin=80] "UML_OCL.level1.encoding.A_2_B"}  \notag \shoveleft{}\\ 
	    \shoveright{} @{thm [break,margin=80] UML_OCL.level1.encoding.A_2_B_def}\\
         \end{multline} 

*}
 text {*
   Finally, we provide accessor functions for the class attributes:
         \begin{multline}
          @{const [long_names] UML_OCL.level1.encoding.A.i} \isasymColon   \notag\\
	     @{typeof [break,margin=80] "UML_OCL.level1.encoding.A.i"}  \notag \shoveleft{}\\ 
	    \shoveright{} @{thm [break,margin=80] UML_OCL.level1.encoding.A.i_def}\\
         \end{multline} 

         \begin{multline}
          @{const [long_names] UML_OCL.level1.encoding.B.j} \isasymColon   \notag\\
	     @{typeof [break,margin=80] "UML_OCL.level1.encoding.B.j"}  \notag \shoveleft{}\\ 
	    \shoveright{} @{thm [break,margin=80] UML_OCL.level1.encoding.B.j_def}\\
         \end{multline} 
*}

subsection {* Proving Properties *} 
   text {* Already on this level, we can prove important properties of our encoding. For 
     example, 
           the following theorem  
           ``@{text UML_OCL.level1.encoding.is_A_univ_implies_OclAny_univ}'' shows, that on the
	   universe types, every object of class \inlineocl{A} is also an object of class
	   \inlineocl{OclAny}: 
	   \begin{multline}\notag
                  @{thm [break,margin=80] UML_OCL.level1.encoding.is_A_univ_implies_OclAny_univ } \\
           \end{multline}
	   Using @{text get_mk_A_id}  and @{text get_mk_A_id} we can convert ``lossless'' between our 
	   class type representation  and universe representation, this is shown by 
           ``@{text UML_OCL.level1.encoding.get_mk_A_id}'':
           \begin{multline}\notag
                  @{thm [break,margin=80] UML_OCL.level1.encoding.get_mk_A_id }  \\
           \end{multline}
           and ``@{text UML_OCL.level1.encoding.mk_get_A_id}'': 
           \begin{multline}\notag
                  @{thm [break,margin=80] UML_OCL.level1.encoding.mk_get_A_id }  \\
           \end{multline}
           and ``@{text UML_OCL.level1.encoding.is_A_univ_implies_is_get}''  shows that 
	   a successfull type test on the universe types also implies a successfull type
	   test on the class types: 
           \begin{multline}\notag
                @{thm [break,margin=80] UML_OCL.level1.encoding.is_A_univ_implies_is_get } \\
           \end{multline}\notag
	   The theorem ``@{text UML_OCL.level1.encoding.is_mk_A}'' shows that the constructor 
	   creates an object that is of 
	   type \inlineocl{A}:
           \begin{multline}\notag
                  @{thm [break,margin=80] UML_OCL.level1.encoding.is_mk_A }  \\
           \end{multline}
	   and also satisfies the test of its super-type, e.g., \inlineocl{OclAny}. This is shown 
	   by the following theorem 
	   ``@{text UML_OCL.level1.encoding.is_OclAny_mk_A}'': 
           \begin{multline}\notag
                  @{thm [break,margin=80] UML_OCL.level1.encoding.is_OclAny_mk_A}  \\
           \end{multline}
	   ``@{text UML_OCL.level1.encoding.cast_A_id}'' shows that we can successfully
	   downcast objects: 
           \begin{multline}\notag
                  @{thm [break,margin=80] UML_OCL.level1.encoding.cast_A_id } \\
           \end{multline}

           Analogous, these properties are also shown for class \inlineocl{B}: \\
           @{text UML_OCL.level1.encoding.is_B_univ_implies_A_univ}: 
	   \begin{multline}\notag
                  @{thm [break,margin=80] UML_OCL.level1.encoding.is_B_univ_implies_A_univ } \\
           \end{multline}
           @{text UML_OCL.level1.encoding.is_B_univ_implies_is_get}: 
           \begin{multline}\notag
                @{thm [break,margin=80] UML_OCL.level1.encoding.is_B_univ_implies_is_get } \\
           \end{multline}\notag
           @{text UML_OCL.level1.encoding.get_mk_B_id}:
           \begin{multline}\notag
                  @{thm [break,margin=80] UML_OCL.level1.encoding.get_mk_B_id }  \\
           \end{multline}
           @{text UML_OCL.level1.encoding.is_mk_B}: 
           \begin{multline}\notag
                  @{thm [break,margin=80] UML_OCL.level1.encoding.is_mk_B }  \\
           \end{multline}
           @{text UML_OCL.level1.encoding.is_A_mk_B}: 
           \begin{multline}\notag
                  @{thm [break,margin=80] UML_OCL.level1.encoding.is_A_mk_B}  \\
           \end{multline}
           @{text UML_OCL.level1.encoding.mk_get_B_id}: 
           \begin{multline}\notag
                  @{thm [break,margin=80] UML_OCL.level1.encoding.mk_get_B_id }  \\
           \end{multline}
           @{text UML_OCL.level1.encoding.cast_B_id}: 
           \begin{multline}\notag
                  @{thm [break,margin=80] UML_OCL.level1.encoding.cast_B_id } \\
           \end{multline}
  
	   Further we show that our definitions are context passing:\\
           @{text UML_OCL.level1.encoding.cp_get_A}: 
	   \begin{multline}\notag
                  @{thm [break,margin=80] UML_OCL.level1.encoding.cp_get_A } \\
           \end{multline}
         @{text UML_OCL.level1.encoding.cp_mk_A}: 
	   \begin{multline}\notag
                  @{thm [break,margin=80] UML_OCL.level1.encoding.cp_mk_A } \\
           \end{multline}
         @{text UML_OCL.level1.encoding.cp_is_A}: 
	   \begin{multline}\notag
                  @{thm [break,margin=80] UML_OCL.level1.encoding.cp_is_A } \\
           \end{multline}
         @{text UML_OCL.level1.encoding.cp_is_univ_A}: 
	   \begin{multline}\notag
                  @{thm [break,margin=80] UML_OCL.level1.encoding.cp_is_univ_A } \\
           \end{multline}

           @{text UML_OCL.level1.encoding.cp_get_B}: 
	   \begin{multline}\notag
                  @{thm [break,margin=80] UML_OCL.level1.encoding.cp_get_B } \\
           \end{multline}
         @{text UML_OCL.level1.encoding.cp_mk_B}: 
	   \begin{multline}\notag
                  @{thm [break,margin=80] UML_OCL.level1.encoding.cp_mk_B } \\
           \end{multline}
         @{text UML_OCL.level1.encoding.cp_is_B}: 
	   \begin{multline}\notag
                  @{thm [break,margin=80] UML_OCL.level1.encoding.cp_is_B } \\
           \end{multline}
         @{text UML_OCL.level1.encoding.cp_is_univ_B}: 
	   \begin{multline}\notag
                  @{thm [break,margin=80] UML_OCL.level1.encoding.cp_is_univ_B } \\
           \end{multline}

 *}


section {*   Encoding Level 2: *}
subsection{* Encoding of \OCL Invariants: Part One *}
text {* Here we bring the data part (i.e., \UML) together with the \OCL specification. The encoding 
  of \OCL formulae is two-staged. First, the formulae are encoding without requiring the 
  definedness  of the occurring path expressions, we extend these formulae later with the required
  definedness constraints. This two-staged process allows us to break cyclic dependencies and thus 
  encode \OCL conservatively. 

  Thus, for class \inlineocl{A} we generate two invariants; The first one describes the 
  initial value of the attribute  \inlineocl{i}
          \begin{multline}
            @{const [long_names] encoding.A.model.hol_invariant.init_i }  \isasymColon       \notag\\
	    @{typeof [break,margin=80] "encoding.A.model.hol_invariant.init_i"}  \notag \shoveleft{}\\ 
	    \shoveright{} @{thm [break,margin=80] encoding.A.model.hol_invariant.init_i_def}  \\
	  \end{multline}
  and the second one describes the user defined invariant:
         \begin{multline}
          @{const [long_names] encoding.A.model.hol_invariant.inv_1 }  \isasymColon       \notag\\
	    @{typeof [break,margin=80] "encoding.A.model.hol_invariant.inv_1"}  \notag \shoveleft{}\\ 
	    \shoveright{} @{thm [break,margin=80] encoding.A.model.hol_invariant.inv_1_def}  \\
	  \end{multline} 
  The overall invariant for class \inlineocl{A} is constructed by conjoining these invariants.
	  \begin{multline}
            @{const [long_names] encoding.A.model.hol_invariant }  \isasymColon       \notag\\
	    @{typeof [break,margin=80] "encoding.A.model.hol_invariant"}  \notag \shoveleft{}\\ 
	    \shoveright{} @{thm [break,margin=80] encoding.A.model.hol_invariant_def}  \\
	  \end{multline}

   Analgous  we are encoding the pre-condition and post-condition of the operation \inlineocl{f}:
          \begin{multline}
            @{const [long_names] encoding.A.f_Integer_Integer.pre1.pre_0}  \isasymColon       \notag\\
	    @{typeof [break,margin=80] "encoding.A.f_Integer_Integer.pre1.pre_0"}  \notag \shoveleft{}\\ 
	    \shoveright{} @{thm [break,margin=80] encoding.A.f_Integer_Integer.pre1.pre_0_def}  \\
	  \end{multline}
          \begin{multline} 
            @{const [long_names] encoding.A.f_Integer_Integer.post1.post_0}  \isasymColon       \notag\\
	    @{typeof [break,margin=80] "encoding.A.f_Integer_Integer.post1.post_0"}  \notag \shoveleft{}\\ 
	    \shoveright{} @{thm [break,margin=80] encoding.A.f_Integer_Integer.post1.post_0_def}  \\
	  \end{multline}
          \begin{multline} 
            @{const [long_names] encoding.A.f_Integer_Integer.pre1}  \isasymColon       \notag\\
	    @{typeof [break,margin=80] "encoding.A.f_Integer_Integer.pre1"}  \notag \shoveleft{}\\ 
	    \shoveright{} @{thm [break,margin=80] encoding.A.f_Integer_Integer.pre1_def}  \\
	  \end{multline}
          \begin{multline} 
            @{const [long_names] encoding.A.f_Integer_Integer.post1}  \isasymColon       \notag\\
	    @{typeof [break,margin=80] "encoding.A.f_Integer_Integer.post1"}  \notag \shoveleft{}\\ 
	    \shoveright{} @{thm [break,margin=80] encoding.A.f_Integer_Integer.post1_def}  \\
	  \end{multline}

  For class  \inlineocl{B} no initial values or invariants are given, thus we generate invariants 
  that are just \inlineocl{true}:
          \begin{multline}
            @{const [long_names] encoding.B.inv.init_j }  \isasymColon       \notag\\
	    @{typeof [break,margin=80] "encoding.B.inv.init_j"}  \notag \shoveleft{}\\ 
	    \shoveright{} @{thm [break,margin=80] encoding.B.inv.init_j_def}  \\
	  \end{multline}
          \begin{multline}
            @{const [long_names] encoding.B.inv.init_j }  \isasymColon       \notag\\
	    @{typeof [break,margin=80] "encoding.B.inv.init_j"}  \notag \shoveleft{}\\ 
	    \shoveright{} @{thm [break,margin=80] encoding.B.inv.init_j_def}  \\
	  \end{multline}
  The overall invariant for class \inlineocl{B} is constructed by conjoining these invariants.
	  \begin{multline}
            @{const [long_names] encoding.B.inv.inv }  \isasymColon       \notag\\
	    @{typeof [break,margin=80] "encoding.B.inv.inv"}  \notag \shoveleft{}\\ 
	    \shoveright{} @{thm [break,margin=80] encoding.B.inv.inv_def}  \\
	  \end{multline}



*}

subsection{* Encoding Of Operations *}
  text {*
         First we define a constant for the operation \inlineocl{f} describing its core semantics
	 specified by the pre-condition and post-condition:
	  \begin{multline}
            @{const [long_names] encoding.A.f }  \isasymColon       \notag\\
	    @{typeof [break,margin=80] "encoding.A.f"}  \notag \shoveleft{}\\ 
	    \shoveright{} @{thm [break,margin=80] encoding.A.f_def}  \\
	  \end{multline}
          For each method, we create a table storing all variants, i.e, overloadingsL
          \begin{multline}
            @{const [long_names] UML_OCL.ops.f_Integer_Integer.tab}  \isasymColon       \notag\\
	    @{typeof [break,margin=80] "UML_OCL.ops.f_Integer_Integer.tab"}  \notag \shoveleft{}\\
	  \end{multline}
	  The entry for \inlineocl{f} in this operation table is defined as follows:
	  \begin{multline}
            @{const [long_names] UML_OCL.ops.f_Integer_Integer.Op}  \isasymColon       \notag\\
	    @{typeof [break,margin=80] "UML_OCL.ops.f_Integer_Integer.Op"}  \notag \shoveleft{}\\ 
	    \shoveright{} @{thm [break,margin=80] UML_OCL.ops.f_Integer_Integer.Op_def}  \\
	  \end{multline}
	  and as  @{text UML_OCL.ops.f_Integer_Integer.overwrite_A}:
	  \begin{multline}
	   @{thm [break,margin=80] UML_OCL.ops.f_Integer_Integer.overwrite_A}  \\
	  \end{multline}
*} 

subsection {* Encoding the Basic \UML part*}
 text {* 
  The encoding of the data model on level 2 follows strictly the known
  schema from level 0 and 1, i.e, we also start two type tests for
  objects, one for testing if a object is of a specific type or subtype
  (i.e., this test is true, if the object is of the given kind):
	   \begin{multline}
          @{const [long_names] encoding.is_A}  \isasymColon      \notag\\
	   @{typeof [break,margin=80] "encoding.is_A"}        \notag \shoveleft{}\\ 
	    \shoveright{} @{thm [break,margin=80] encoding.is_A_def} \\
         \end{multline}
and one for testing if a object is exactly the requested type 
         \begin{multline}
	  @{const [long_names] encoding.isType_A}  \isasymColon   \notag\\
	   @{typeof [break,margin=80] "encoding.isType_A"}    \notag \shoveleft{}\\ 
	    \shoveright{} @{thm [break,margin=80] encoding.isType_A_def}\\
         \end{multline}
Further, we define a type for universe instances:
         \begin{multline}
	  @{const [long_names] encoding.is_A_univ}  \isasymColon  \notag\\
	   @{typeof [break,margin=80] "encoding.is_A_univ"}   \notag \shoveleft{}\\ 
	    \shoveright{} @{thm [break,margin=80] encoding.is_A_univ_def}\\
         \end{multline}
	 Analogous we define for the inherited class  \inlineocl{B}:
	   \begin{multline}
          @{const [long_names] encoding.is_B}  \isasymColon      \notag\\
	   @{typeof [break,margin=80] "encoding.is_B"}        \notag \shoveleft{}\\ 
	    \shoveright{} @{thm [break,margin=80] encoding.is_B_def} \\
         \end{multline}
         \begin{multline}
	  @{const [long_names] encoding.isType_B}  \isasymColon   \notag\\
	   @{typeof [break,margin=80] "encoding.isType_B"}    \notag \shoveleft{}\\ 
	    \shoveright{} @{thm [break,margin=80] encoding.isType_B_def}\\
         \end{multline}
         \begin{multline}
	  @{const [long_names] encoding.is_B_univ}  \isasymColon  \notag\\
	   @{typeof [break,margin=80] "encoding.is_B_univ"}   \notag \shoveleft{}\\ 
	    \shoveright{} @{thm [break,margin=80] encoding.is_B_univ_def}\\
         \end{multline}
*}
text {*
       We define a functions for embeding a class into the corresponding universe
          \begin{multline}
            @{const [long_names] encoding.mk_A}  \isasymColon       \notag\\
	    @{typeof [break,margin=80] "encoding.mk_A"}  \notag \shoveleft{}\\ 
	    \shoveright{} @{thm [break,margin=80] encoding.mk_A_def}  \\
	  \end{multline}
	  and one for extracting an object out of a given universe:
	  \begin{multline}
          @{const [long_names] encoding.get_A}  \isasymColon      \notag\\
	   @{typeof [break,margin=80] "encoding.get_A"}      \notag \shoveleft{}\\ 
	    \shoveright{} @{thm [break,margin=80] encoding.get_A_def}\\
	   \end{multline}
	   These two functions can also be understood as conversion between object types
	   and universe types.

 	 Analogous we define for the inherited class  \inlineocl{B}:
          \begin{multline}
            @{const [long_names] encoding.mk_B}  \isasymColon       \notag\\
	    @{typeof [break,margin=80] "encoding.mk_B"}  \notag \shoveleft{}\\ 
	    \shoveright{} @{thm [break,margin=80] encoding.mk_B_def}  \\
	  \end{multline}
	  \begin{multline}
          @{const [long_names] encoding.get_B}  \isasymColon      \notag\\
	   @{typeof [break,margin=80] "encoding.get_B"}      \notag \shoveleft{}\\ 
	    \shoveright{} @{thm [break,margin=80] encoding.get_B_def}\\
	   \end{multline}
*}
text {*
        For converting along type hierarchies (i.e. down-casting and up-casting) we
	provide for each class two functions for the conversion to the direct predecessor 
	and direct successor in the class hierarchy:
        \begin{multline}
          @{const [long_names] encoding.A_2_OclAny}  \isasymColon  \notag\\
	   @{typeof [break,margin=80] "encoding.A_2_OclAny"}   \notag \shoveleft{}\\ 
	    \shoveright{} @{thm [break,margin=80] encoding.A_2_OclAny_def}\\
         \end{multline}
         \begin{multline}
          @{const [long_names] encoding.OclAny_2_A} \isasymColon   \notag\\
	     @{typeof [break,margin=80] "encoding.OclAny_2_A"}  \notag \shoveleft{}\\ 
	    \shoveright{} @{thm [break,margin=80] encoding.OclAny_2_A_def}\\
         \end{multline}
	 Analogous we define for the inherited class  \inlineocl{B}:
         \begin{multline}
          @{const [long_names] encoding.B_2_A}  \isasymColon  \notag\\
	   @{typeof [break,margin=80] "encoding.B_2_A"}   \notag \shoveleft{}\\ 
	    \shoveright{} @{thm [break,margin=80] encoding.B_2_A_def}\\
         \end{multline}
         \begin{multline}
          @{const [long_names] encoding.A_2_B} \isasymColon   \notag\\
	     @{typeof [break,margin=80] "encoding.A_2_B"}  \notag \shoveleft{}\\ 
	    \shoveright{} @{thm [break,margin=80] encoding.A_2_B_def}\\
         \end{multline} 

*}
 text {*
   Finally, we provide accesor functions for the class attributes:
         \begin{multline}
          @{const [long_names] encoding.A.i} \isasymColon   \notag\\
	     @{typeof [break,margin=80] "encoding.A.i"}  \notag \shoveleft{}\\ 
	    \shoveright{} @{thm [break,margin=80] encoding.A.i_def}\\
         \end{multline} 

         \begin{multline}
          @{const [long_names] encoding.B.j} \isasymColon   \notag\\
	     @{typeof [break,margin=80] "encoding.B.j"}  \notag \shoveleft{}\\ 
	    \shoveright{} @{thm [break,margin=80] encoding.B.j_def}\\
         \end{multline} 
*}

subsection {* Defining \inlineocl{oclAsType} and \inlineocl{allInstances} *}
 
text {* The constants for  \inlineocl{oclAsType} and \inlineocl{allInstances} 
  must be defined for each class separately (using overloading). We provide 
  for each class two definitions for  \inlineocl{oclAsType}: one for the 
  actual state (``@{text UML_OCL.univ.encoding.A.OclAllInstancesencoding_A}'')   
  \begin{multline}\notag
    @{thm [break,margin=80] UML_OCL.univ.encoding.A.OclAllInstancesencoding_A_def}  
  \end{multline} 
  and one for the previous state (``@{text UML_OCL.univ.encoding.A.OclAllInstancesencoding_AAtpre}''):   
  \begin{multline}\notag
    @{thm [break,margin=80] UML_OCL.univ.encoding.A.OclAllInstancesencoding_AAtpre_def}  
  \end{multline}  
  Analgous for class \inlineocl{B} we define 
           ``@{text UML_OCL.univ.encoding.B.OclAllInstancesencoding_B}'':
         \begin{multline}\notag
	   @{thm [break,margin=80] UML_OCL.univ.encoding.B.OclAllInstancesencoding_B_def }\\
         \end{multline} 
         and ''@{text UML_OCL.univ.encoding.B.OclAllInstancesencoding_BAtpre}'':
         \begin{multline}\notag
	   @{thm [break,margin=80] UML_OCL.univ.encoding.B.OclAllInstancesencoding_BAtpre_def }\\
         \end{multline} 

*}

text {* In a similar way, we have to overload the definition of \inlineocl{oclAsType}, based
  on the already defined conversion for adjacent classes. For class \inlineocl{A} the 
following cases must be considered along the inheritance hierarchy: The downcast from
 \inlineocl{OclAny} ``@{text UML_OCL.univ.encoding.A.from_OclAny}'':
	   \begin{equation}\notag
	  @{thm [break,margin=80] UML_OCL.univ.encoding.A.from_OclAny_def}  
	\end{equation} 
	the upcast to  \inlineocl{OclAny} ``@{text UML_OCL.univ.encoding.A.to_OclAny}'': 
	   \begin{equation}\notag
	  @{thm [break,margin=80] UML_OCL.univ.encoding.A.to_OclAny_def}  
	\end{equation} 
	and the identity ``@{text UML_OCL.univ.encoding.A.to_A}'': 
	   \begin{equation}\notag
	     @{thm [break,margin=80] UML_OCL.univ.encoding.A.to_A_def}  
	   \end{equation} 
	
	For class \inlineocl{B} more cases have to be considered, i.e.,
	the downcast from \inlineocl{A} ``@{text UML_OCL.univ.encoding.B.from_A}'':
        \begin{equation}\notag
	   @{thm [break,margin=80] UML_OCL.univ.encoding.B.from_A_def }
         \end{equation} 
	 the downcast from  \inlineocl{OclAny} ``@{text UML_OCL.univ.encoding.B.from_OclAny}'':
	 \begin{equation}\notag
	   @{thm [break,margin=80] UML_OCL.univ.encoding.B.from_OclAny_def }
         \end{equation} 
	 the upcast to \inlineocl{OclAny} ``@{text UML_OCL.univ.encoding.B.to_OclAny}'':
	 \begin{equation}\notag
	   @{thm [break,margin=80] UML_OCL.univ.encoding.B.to_OclAny_def }
         \end{equation}
	 the upcast to \inlineocl{A} ``@{text UML_OCL.univ.encoding.B.to_A}'':
         \begin{equation}\notag
	   @{thm [break,margin=80]  UML_OCL.univ.encoding.B.to_A_def }
         \end{equation} 
	 and the identity ``@{text UML_OCL.univ.encoding.B.to_B}'':
         \begin{equation}\notag
	   @{thm [break,margin=80] UML_OCL.univ.encoding.B.to_B_def }
         \end{equation} 
	 Optional, the encoder generates all definitions that are orthogonal to the 
	 class hierarchy. On the one hand, they are defined to be \inlineocl{OclUndefined} according 
	 to the \OCL standard, on the other hand, they violate a well-formed constraint.
*}

subsection{* Encoding of \OCL Invariants: Part Two *}
text {* 
  Finally, we generate the the final invariant, e.g.,  for class \inlineocl{A} we generate:
          \begin{multline}
            @{const [long_names] encoding.A.inv.init_i }  \isasymColon       \notag\\
	    @{typeof [break,margin=80] "encoding.A.inv.init_i"}  \notag \shoveleft{}\\ 
	    \shoveright{} @{thm [break,margin=80] encoding.A.inv.init_i_def}  \\
	  \end{multline}
  and the second one describes the user defined invariant:
         \begin{multline}
          @{const [long_names] encoding.A.inv.inv_1 }  \isasymColon       \notag\\
	    @{typeof [break,margin=80] "encoding.A.inv.inv_1"}  \notag \shoveleft{}\\ 
	    \shoveright{} @{thm [break,margin=80] encoding.A.inv.inv_1_def}  \\
	  \end{multline} 
  The overall invariant for class \inlineocl{A} is constructed by conjoining these invariants.
	  \begin{multline}
            @{const [long_names] encoding.A.inv }  \isasymColon       \notag\\
	    @{typeof [break,margin=80] "encoding.A.inv"}  \notag \shoveleft{}\\ 
	    \shoveright{} @{thm [break,margin=80] encoding.A.inv_def}  \\
	  \end{multline}

   Analogous  we are encoding the pre-condition and post-condition of the operation \inlineocl{f}:
          \begin{multline}
            @{const [long_names] encoding.A.f_Integer_Integer.pre1.pre_0}  \isasymColon       \notag\\
	    @{typeof [break,margin=80] "encoding.A.f_Integer_Integer.pre1.pre_0"}  \notag \shoveleft{}\\ 
	    \shoveright{} @{thm [break,margin=80] encoding.A.f_Integer_Integer.pre1.pre_0_def}  \\
	  \end{multline}
          \begin{multline} 
            @{const [long_names] encoding.A.f_Integer_Integer.post1.post_0}  \isasymColon       \notag\\
	    @{typeof [break,margin=80] "encoding.A.f_Integer_Integer.post1.post_0"}  \notag \shoveleft{}\\ 
	    \shoveright{} @{thm [break,margin=80] encoding.A.f_Integer_Integer.post1.post_0_def}  \\
	  \end{multline}
          \begin{multline} 
            @{const [long_names] encoding.A.f_Integer_Integer.pre1}  \isasymColon       \notag\\
	    @{typeof [break,margin=80] "encoding.A.f_Integer_Integer.pre1"}  \notag \shoveleft{}\\ 
	    \shoveright{} @{thm [break,margin=80] encoding.A.f_Integer_Integer.pre1_def}  \\
	  \end{multline}
          \begin{multline} 
            @{const [long_names] encoding.A.f_Integer_Integer.post1}  \isasymColon       \notag\\
	    @{typeof [break,margin=80] "encoding.A.f_Integer_Integer.post1"}  \notag \shoveleft{}\\ 
	    \shoveright{} @{thm [break,margin=80] encoding.A.f_Integer_Integer.post1_def}  \\
	  \end{multline}

  For class  \inlineocl{B} no initial values or invariants are given, thus we generate invariants 
  that are just \inlineocl{true}:
          \begin{multline}
            @{const [long_names] encoding.B.inv.init_j }  \isasymColon       \notag\\
	    @{typeof [break,margin=80] "encoding.B.inv.init_j"}  \notag \shoveleft{}\\ 
	    \shoveright{} @{thm [break,margin=80] encoding.B.inv.init_j_def}  \\
	  \end{multline}
          \begin{multline}
            @{const [long_names] encoding.B.inv.init_j }  \isasymColon       \notag\\
	    @{typeof [break,margin=80] "encoding.B.inv.init_j"}  \notag \shoveleft{}\\ 
	    \shoveright{} @{thm [break,margin=80] encoding.B.inv.init_j_def}  \\
	  \end{multline}
  The overall invariant for class \inlineocl{B} is constructed by conjoining these invariants.
	  \begin{multline}
            @{const [long_names] encoding.B.inv }  \isasymColon       \notag\\
	    @{typeof [break,margin=80] "encoding.B.inv"}  \notag \shoveleft{}\\ 
	    \shoveright{} @{thm [break,margin=80] encoding.B.inv_def}  \\
	  \end{multline}



*}
 
text_raw {*
  \let\isanewline\oldisanewline
  \let\isa\oldisa
 *}
(*<*)
end
(*>*)


