(*****************************************************************************
 *          HOL-OCL - a shallow embedding of OCL in Isabelle/HOL              
 *                                                                            
 * OCL_State.thy - State accessing operations as defined in the standard.
 * Copyright (C) 2001-2006 Achim D. Brucker <brucker@inf.ethz.ch>   
 *                         Burkhart Wolff   <bwolff@inf.ethz.ch>    
 *                                                                            
 * This file is part of HOL-OCL.                                              
 *                                                                            
 * HOL-OCL is free software; you can redistribute it and/or modify it under   
 * the terms of the GNU General Public License as published by the Free       
 * Software Foundation; either version 2 of the License, or (at your option)  
 * any later version.                                                         
 *                                                                            
 * HOL-OCL is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS 
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more 
 * details.                                                              
 *                                                                            
 * You should have received a copy of the GNU General Public License along    
 * with this program; if not, write to the Free Software Foundation, Inc.,    
 * 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.                  
 ******************************************************************************)

header {* \OCL State *}

theory OCL_State  
imports 
  "$HOLOCL_HOME/src/library/OclAny/OCL_OclAny" 
begin
text_raw{* \label{thy:OCL_State} *}
types 
      'a state  = "oid \<rightharpoonup> 'a"
      's St     = "'s state \<times> 's state"  
      ('a,'t) V = " ('a St, 't) VAL" 

translations
      "a state"  <= (type) "oid \<Rightarrow> a option"   
      "a St"     <= (type) "(a state) \<times> (s state)" 
      "(a,t) V"  <= (type) "a St \<Rightarrow>  t"

text{* To sum up, the \HOL types assigned to the key \OCL
  types of the OCL standard are a context (st) indexed
  family of types. Recall: the contexts are a state relation
  or a Kripke-Structure over states over the universe.

  Provided that we have a representation of class types
  generated from an \UML class diagram, we can therefore
  translate any \OCL type into an HOL-type achieving
  type soundness on the \HOL-level.
 *}

section{* Checking Freshness of Object Instances in the State.*}
constdefs
  OclIsNew     :: "('a U St,'b::bot)VAL  \<Rightarrow> ('a U St) Boolean"  
 "OclIsNew obj \<equiv> (\<lambda>(s,s'). \<HolOclLiftLeft>level0.mk_OclAny (((obj) ->oclAsType(OclAny)) (s,s')) \<notin> ran s 
                    \<and> level0.mk_OclAny (((obj) ->oclAsType(OclAny) ) (s,s')) \<in> ran s'\<HolOclLiftRight>)"


text{* 
  The \OCL encoder must generate a coercion in order
  to make this operation applicable to any object of some class-type *}   

section{* Selecting the Objects living in a State.*}
consts 
   OclAllInstances      :: "('a U St,'b)VAL \<Rightarrow> ('a U St,'b Set_0)VAL"  
   OclAllInstancesAtpre :: "('a U St,'b)VAL \<Rightarrow> ('a U St,'b Set_0)VAL"  

defs
  OclAllInstancesString_def[simp]:
  "OclAllInstances (self::('a U St,String_0)VAL) \<equiv> OclUndefined"

  OclAllInstancesPreString_def[simp]:
  "OclAllInstancesAtpre (self::('a U St,String_0)VAL) \<equiv> OclUndefined"  

defs
  OclAllInstancesInteger_def[simp]:
  "OclAllInstances (self::('a U St,Integer_0)VAL) \<equiv> OclUndefined"

  OclAllInstancesPreInteger_def[simp]:
  "OclAllInstancesAtpre (self::('a U St,Integer_0)VAL) \<equiv> OclUndefined"  

defs
  OclAllInstancesReal_def[simp]:
  "OclAllInstances (self::('a U St,Real_0)VAL) \<equiv> OclUndefined"

  OclAllInstancesPreReal_def[simp]:
  "OclAllInstancesAtpre (self::('a U St,Real_0)VAL) \<equiv> OclUndefined"  


defs
  OclAllInstancesBoolean_def[simp]:
  "OclAllInstances (self::('a U St,Boolean_0)VAL) \<equiv> Boolean_0"

  OclAllInstancesPreBoolean_def[simp]:
  "OclAllInstancesAtpre (self::('a U St,Boolean_0)VAL) \<equiv> Boolean_0"  

text{* And now the general case for all "objects", i.e. all instances
       of subclass OclAny: *}
defs
  OclAllInstancesOclAny_def:
  "OclAllInstances (self::('a U St, 'a OclAny_0)VAL) \<equiv> 
   \<lambda>(s,s'). Abs_Set_0 (\<HolOclLiftLeft>level0.get_OclAny ` ran s'\<HolOclLiftRight>)"

  OclAllInstancesPreOclAny_def:
  "OclAllInstancesAtpre (self::('a U St, 'a OclAny_0)VAL) \<equiv> 
   \<lambda>(s,s'). Abs_Set_0 (\<HolOclLiftLeft>level0.get_OclAny ` ran s\<HolOclLiftRight>)"

section{* The \inlineocl{modifiedOnly()}-clause and its Infrastructure*}
constdefs OclOidOf :: "'a U state \<Rightarrow> 'a OclAny_0 \<Rightarrow> oid set"
         "OclOidOf s X \<equiv> ({x . (s x) = Some(OCL_OclAny.mk_OclAny X)})"

text{* The $\HolOclSet{\oid}$ is a tribute to non-referential universes
   where there may be different object identifiers in the store for
   the same object. The redtriction to @{text "'\<alpha> OclAny_0"} assures
   that the projection is only applicable to objects. *}  
constdefs OclModifiedOnly :: "(('a U St,'a OclAny_0) Set) \<Rightarrow> ('a U St) Boolean"    
         "OclModifiedOnly X \<equiv> (\<lambda>(s,s'). 
            \<HolOclLiftLeft>\<forall>id \<in> -(\<Union> (OclOidOf s)`\<HolOclDropLeft>Rep_Set_0(X(s,s'))\<HolOclDropRight>). s id = s' id\<HolOclLiftRight>)"
end
