(*****************************************************************************
 *          HOL-OCL - a shallow embedding of OCL in Isabelle/HOL              
 *                                                                            
 * UML_OCL.thy -
 * Copyright (C) 2001-2006 Achim D. Brucker <brucker@inf.ethz.ch>   
 *                         Burkhart Wolff   <bwolff@inf.ethz.ch>    
 *                                                                            
 * This file is part of HOL-OCL.                                              
 *                                                                            
 * HOL-OCL is free software; you can redistribute it and/or modify it under   
 * the terms of the GNU General Public License as published by the Free       
 * Software Foundation; either version 2 of the License, or (at your option)  
 * any later version.                                                         
 *                                                                            
 * HOL-OCL is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS 
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more 
 * details.                                                              
 *                                                                            
 * You should have received a copy of the GNU General Public License along    
 * with this program; if not, write to the Free Software Foundation, Inc.,    
 * 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.                  
 ******************************************************************************)
header {* The Object Constraint Language: \OCL *}

theory UML_OCL 
imports 
  "$HOLOCL_HOME/src/OCL_State"
  "$HOLOCL_HOME/src/OCL_Operation"   
begin
text_raw{* \label{thy:UML_OCL} *}

section{* The Theory-Morpher *} 
text{* 
  In the theory file, this section contains an \acs{sml} program
  implementing a theory morphism from \HOL to \holocl. It filters all
  combinator-based definitions in \holocl, contructs a one-to-one
  constant symbol based signature morphism between the \HOL and the
  \holocl theory (modulo argument permutation), uses this signature
  morphism to translate \HOL theorems to \holocl proof goals and
  attempts to prove them with a generic case split tactic. For
  conceptual details, see~\cite{brucker.ea:embedding:2003}.  
*}
(*<*)
ML{*
(******************************************************************************)
(*                                                                            *)
(*  Term Interface                                                            *)
(*                                                                            *)
(******************************************************************************)


fun mk_global_valid X = 
    Const("OCL_Logic_core.OclValid",dummyT) $ X 

fun mk_local_valid X ctxt = 
    Const("OCL_Logic_core.OclLocalValid",dummyT) $ X $ ctxt

fun mk_LJE A B ctxt =
    Const("op =",dummyT) 
    $ (mk_local_valid A ctxt) 
    $ (mk_local_valid B ctxt)

fun mk_GE A B ctxt =
    Const("op =",dummyT) 
    $ (A $ ctxt) 
    $ (B $ ctxt)


fun err_dups names = 
    error ("Duplicate entries in thy_morpher: " ^ commas_quote names); 

fun store_thm2 p (name,thm) = 
    let val clean_dot = String.map 
        (fn x => if x = #"." then #"_" else x)
    in bind_thm(clean_dot((p())^name),thm) end;

fun add_term_consts' (t, cnst: term list) = case t of
    Const   _ => OrdList.insert Term.term_ord t cnst
  | Abs (_,_,body) => add_term_consts'(body,cnst)
  | f$t =>  add_term_consts'(f, add_term_consts'(t, cnst))
  | _ => cnst;


fun consts_of t = map dest_Const (add_term_consts'(t,[]));
fun frees_of t = map dest_Free (term_frees t)


fun subtract xs xt = filter (fn (x,_) => forall(fn y => not(x = y)) xt)(xs)

fun subtract' xs xt = filter (fn (x,t) => forall(fn (y,t') => not(x = y andalso
                                                                  Type.raw_instance(t',t))) xt)
                             (xs)


fun args (Bound x) = [x]
   |args (s $ t) = (args s) @ (args t)
   |args (Abs(_,_,t)) = map (fn x => x + 1) (args t)
   |args _ = []


(* PRE : t does not contain Bound's *)
fun mk_rev t = Abs("x",dummyT,Abs("y",dummyT,t $ (Bound 1) $ (Bound 0)))


(* PRE : t does not contain Bound's *)
fun mk_param [] t    = t
   |mk_param [0] t   = t
   |mk_param [1,0]t  = t
   |mk_param [2,1,0]t= t 
   |mk_param M t = 
    let fun H []     = list_comb(t, map Bound M)
           |H (a::S) = Abs("",dummyT, H S)
    in  H M end

(******************************************************************************)
(*                                                                            *)
(*  Computation of the "Core" of the Signature Morphism                       *)
(*                                                                            *)
(******************************************************************************)


(* Serious limitation of the current approach: the thy-morhpism is
   based on a signature morphism that must be representable as
   one-to-one relation on constant symbols.

   This rules out many examples where some function compositions are
   needed: e.g. including in OCL_OrderedSet, where a an @ *and*
   an injection into lists is needed \<dots> 
 
   A slight relaxation is an assignment to a constant and
   a lambda-expression (allowing for parameter permutations.)
   However, this extension is probably not worth the effort\<dots>
  *)



val SEM_COM = ["op o", "Let","Lifting.strictify",
               "Lifting.lift1","Lifting.lift2","Lifting.lift3",
               "Lifting.up.lift", "Lifting.up.down", 
               "Lifting.drop", "Lifting.UU",
               "OCL_Set_type.Abs_Set_0", "OCL_Set_type.Rep_Set_0",
               "OCL_Sequence_type.Abs_Sequence_0", 
               "OCL_Sequence_type.Rep_Sequence_0",
               "OCL_OrderedSet_type.Abs_OrderedSet_0", 
               "OCL_OrderedSet_type.Rep_OrderedSet_0",
               "OCL_Bag_type.Abs_Bag_0", "OCL_Bag_type.Rep_Bag_0",
               "IntDef.int", "IntDef.nat"] (* considered as datatype-adaption. *) 



fun find_thym_pair (name,thm) = (* (cterm,cterm,thm) option *)
       case concl_of thm of
          (Const("==",_ ) $ lhs $
            (Const ("Lifting.lift3", _) $
               (Const("Lifting.strictify",_) $ Abs(_,_, 
               (Const("Lifting.strictify",_) $ Abs(_,_, 
                Const("Lifting.strictify",_) $ Abs(_,_,rhs))))))) 
           =>(case subtract (consts_of rhs) SEM_COM of
                [t] => SOME ((lhs,Const(t)),(name,thm),args rhs)
               | _  => NONE)
       | (Const("==",_ ) $ lhs $
            (Const ("Lifting.lift2", _) $
               (Const("Lifting.strictify",_) $ Abs(_,_, 
                Const("Lifting.strictify",_) $ Abs(_,_,rhs))))) 
           =>(case subtract (consts_of rhs) SEM_COM of
                [t] => SOME ((lhs,Const(t)),(name,thm), args rhs)
               | _  => NONE)
       | (Const("==",_ ) $ lhs $
            (Const ("Lifting.lift1", _) $ 
              (Const("Lifting.strictify",_) $ Abs(_,_,rhs)))) 
           => (case subtract (consts_of rhs) SEM_COM of
                 [t] => SOME ((lhs,Const(t)),(name,thm), args rhs)
               | _   => NONE)
       | (Const("==",_ ) $ lhs $
            (Const ("Lifting.lift0", _) $ rhs)) 
           => (case subtract (consts_of rhs) SEM_COM of
                 [t] => SOME ((lhs,Const(t)),(name,thm), args rhs)
               | _   => NONE)
       | _ => NONE;


fun core thy =
       Library.mapfilter (find_thym_pair) (axioms_of thy);


*}

text{* Test so far, for one single theory: *}
ML{* core (theory"OCL_Sequence") *} 

ML{*


(******************************************************************************)
(*                                                                            *)
(*  Conversion                                                                *)
(*                                                                            *)
(******************************************************************************)


val tauT = Free("\<tau>",dummyT)

fun gen_GE lhs rhs =
    (Const ("Trueprop",HOLogic.boolT --> propT) $ (Const ("op =",dummyT) $ (lhs $ tauT) $ (rhs $ tauT)))

fun gen_LJ prop =
    (Const ("Trueprop",HOLogic.boolT --> propT) $ 
           (Const("OCL_Logic_core.OclLocalValid", dummyT) $ tauT $ prop))

fun gen_LJ_isdef name =
    gen_LJ (Const("OCL_Boolean.OclIsDefined",dummyT) $ Free(name,dummyT))

fun gen_UC lhs rhs =
    (Const ("Trueprop",HOLogic.boolT --> propT) $ (Const ("op =",dummyT) $ lhs $ rhs))



(* an equation lhs = rhs is called regular, iff all free variables in lhs 
   also occur in rhs. Therefore, the irregular free variables are the
   free variables that occur in lhs but not in rhs.
 *)
fun get_irregular_free t1 t2 =  (frees_of t1) \\ (frees_of t2)

fun conv lkup (Free(s,_))    = [Free(s,dummyT)]
   |conv lkup (Var((s,_),_)) = [Free(s,dummyT)]
   |conv lkup (Const(s,t))   = (case lkup(s,t) of
                                 [] => [Const(s,dummyT)]
                               | S  => S)
   |conv lkup (Abs(s,_,t))   = map(fn x => Abs(s,dummyT,x)) (conv lkup t)
   |conv lkup (s $ t)        = flat(map(fn x => map(fn y => x $ y)(conv lkup t))
                                       (conv lkup s))
   |conv lkup x              = [x](* Bound variable *)


fun conv_thm conv (ff as (Const ("==>",t) $ A $ B)) = 
             let val (prems, concl) = Logic.strip_horn ff
                 fun aux(Const("Trueprop",_)$(Const("op =",_)$lhs$rhs))=
                              (flat(map(fn lhs'=>map(gen_GE lhs')(conv rhs))(conv lhs)))
                    |aux(Const("Trueprop",_)$pred) = map gen_LJ (conv pred)
                 fun H [] R = R
                    |H (A::S) R =  flat(map (fn c => map (fn p=>implies$p$c) A) 
                                            (H S R))
             in H(map aux prems)
                 (case concl of
                   (Const ("Trueprop",t) $ 
                       (Const ("op =",_) $ lhs $ rhs)) => 
                          let val irf = get_irregular_free lhs rhs
                              val irf_constraints = map (fn(n,_) =>gen_LJ_isdef n) irf
                          in  map (fn eq => Logic.list_implies(irf_constraints, eq))
                                  (flat(map (fn lhs' => map (gen_GE lhs') (conv rhs)) 
                                            (conv lhs))) 
                          end
                 | (Const ("Trueprop",t) $ pred) => map gen_LJ (conv pred)
                 | _ => [])
                handle _ => [] (* aux may fail *)
             end

   |conv_thm conv (Const ("Trueprop",t) $ 
                  (Const ("op =",_) $ lhs $ rhs))=
             let val irf = get_irregular_free lhs rhs
                 val irf_constraints = map (fn(n,_) =>gen_LJ_isdef n) irf
             in  if null irf 
                 then flat(map (fn lhs' => map (gen_UC lhs') (conv rhs)) (conv lhs)) 
                 else map (fn eq => Logic.list_implies(irf_constraints, eq))
                          (flat(map (fn lhs' => map (gen_GE lhs') (conv rhs)) 
                                    (conv lhs)))
             end
    
   |conv_thm conv (Const ("Trueprop",_) $ A)=
             map gen_LJ (conv A)

   |conv_thm conv _ = []


fun convert thy subst (name,thm) =
    let fun lookup(s,t) = Library.mapfilter 
                                    (fn ((c1,Const(_,t')),((name,def),arg_sig)) => 
                                        if Type.raw_instance(t,t') 
                                        then SOME (mk_param arg_sig c1)
                                        else NONE)
                                    (Symtab.lookup_multi subst s)

	val groundTVars  = map_term_types (Type.unvarifyT);

        val cc = map groundTVars (conv_thm (conv lookup) (term_of(cprop_of thm)));
        
        val errorfn_save = (!error_fn); 
                
        fun type_inference ct   = 
            let val _ = (error_fn:= (fn _ => ()))  
                                             (* type_inference: shut up *)                  
                val res = ((SOME(fst(Sign.infer_types (Sign.pp thy) (thy) 
				          (K NONE) (K NONE) [] false ([ct],propT))))
                           handle _ => NONE)
                                      
                val _ = (error_fn := errorfn_save) (* type_inference: speak up again *)
            in  res end 
        val ccc = Library.mapfilter type_inference cc
     in map (fn x => (name,cterm_of (sign_of thy) x)) ccc
     end;
 
*}

(*
ML {* use "config.sml" *}
*)

ML{*

(******************************************************************************)
(*                                                                            *)
(*  Generic Proof Engine                                                      *)
(*                                                                            *)
(******************************************************************************)

val DEF_exhausts  = case config.collectiontype of
			config.unsmashed => [thm "DEF_X_up"]
		      | config.smashed   => [thm"DEF_X_up",
					     thm"DEF_X_Set",
					     thm"DEF_X_Sequence",
                                             thm"DEF_X_Bag"];

val congr_inv_aux = case config.collectiontype of
			config.unsmashed => []
		      | config.smashed   => [thm"Abs_Set_0_inverse",
					     thm"DEF_Abs_Set",
					     thm"Abs_Sequence_0_inverse",
					     thm"DEF_Abs_Sequence",
                                             thm"Abs_Bag_0_inverse",
					     thm"DEF_Abs_Bag"
                                            ]

val congr_intros  = case config.collectiontype of
			 config.unsmashed => []
		       | config.smashed   => 
			 [((thm"inj_on_Abs_Set") RS inj_on_iff) RS iffD2,
                          ((thm"inj_on_Abs_Sequence") RS inj_on_iff) RS iffD2,
                          ((thm"inj_on_Abs_Bag_0") RS inj_on_iff) RS iffD2
                         ];
                       (* do not use as rewrites !!! *)
val congr_charns  = case config.collectiontype of
			 config.unsmashed => []
		       | config.smashed   => 
			 [(thm"smashed_set_charn") RS sym,
                          (thm"smashed_sequence_charn") RS sym,
                          (thm"smashed_bag_charn") RS sym
                         ];



fun eq_lifter_prover thy thms defs (name,goal) = 
    let val thyname  = hd(String.tokens(fn x => x = #".")(name))
        val core_thms = filter(fn (x,_) => x = name) thms 
        val v        = distinct(frees_of(term_of goal))
        val simpset  = simpset_of thy
        val claset   = claset_of thy
        val fnm      = map (fn (X,_)=> "DEF("^X^" St)") v
        val case_tacs= map (fn X=> case_tac X 1) fnm
    in  SOME(name,prove_goalw_cterm defs goal
                  (fn prems => [ cut_facts_tac prems 1,
                                 ALLGOALS(asm_full_simp_tac 
                                             (simpset addsimps [
                                                  thm"localValidDefined2sem",
                                                  thm"localValidUnDefined2sem",
                                                  thm"localValid2sem"])),
                                 ALLGOALS(asm_full_simp_tac 
                                             (simpset addsimps [thm"localValid2sem"])),
                                 TRY(rtac ext 1)]
                               @ case_tacs @
                               [ ALLGOALS(asm_full_simp_tac 
                                             (simpset addsimps [thm"not_DEF_X"])),
                                 ALLGOALS(asm_full_simp_tac 
                                             (HOL_ss addsimps DEF_exhausts)),
                                 auto_tac (claset  addSIs congr_intros,
                                           simpset addsimps (congr_inv_aux @
                                                               congr_charns)),
                                 TRY(FIRST(map (fn (_,thm) => rtac thm 1) core_thms)), 
                                 ALLGOALS(asm_full_simp_tac simpset)]))
        handle _ => NONE
    end;


*}

ML{*
(******************************************************************************)
(*                                                                            *)
(*  Control                                                                   *)
(*                                                                            *)
(******************************************************************************)

val state = ref   {trace_morpher = true,
                   thms          = [] : (string * Thm.thm) list,
                   liftables     = [] : (string * Thm.thm) list,
                   convertables  = [] : (string * Thm.cterm) list,
                   proven_lifts  = [] : (string * Thm.thm) list} 


fun lift root_thy core_E thms =
    let fun ran core_E   = map (fn ((_,y),_,_)=> y) core_E;
        val the_logical_connectives = ["Trueprop","op =", "Lifting.DEF",
	                               "All", "==>"]
        val _      = (writeln("\n Computing the Core ... ");
                      writeln(  " ====================== \n"));

	val core_E_M     = filter (fn ((_,Const(s,_)),_,_) => 
	                               forall(fn x => not(x = s))
				             (the_logical_connectives))
			          (core_E)

        val the_ran_core_E_M'       = map dest_Const (ran core_E_M)

        fun is_liftable (name,thm)  = 
                           let val cc  = consts_of(term_of(cprop_of thm)) 
                               val cc1 = subtract cc the_logical_connectives
                               val cc2 = subtract' cc1 the_ran_core_E_M'
                           in  not(null(cc1)) andalso null(cc2) end

        fun inst thm(n,t)= (let val h as (_::_) = (term_tvars(concl_of thm))
                                val tt = ctyp_of (theory_of_thm thm)
                                fun convT (x as (s,["HOL.type"])) = raise Match
                                   |convT (x as (s,_)) = (tt(TVar x), tt(t))
                            in  SOME(n,instantiate (map convT h,[]) thm) end
                            handle _ => NONE);

        fun instantiate_classes (name,thm) = 
                            (case Library.mapfilter 
                                       (inst thm) 
                                       [(name^"_int",HOLogic.intT),
                                        (name^"_real",HOLogic.realT)] of
                               [] => [(name,thm)]
                             | S  =>  S);  

        val _      = (writeln("\n Computing the Liftables ... ");
                      writeln(  " =========================== \n"));

        val liftables    = flat(map instantiate_classes
                                    (filter is_liftable thms));

        fun regrp ((x,y as Const(s,t)),a,parms) = (s,((x,y),(a,parms)))
    
        val subst        = Symtab.make_multi(map regrp core_E_M)
                           (* subst is the inversion of the core mapping.
                              Since the core of the morphism is *NOT*
                              injective in general, the inversion may be one-to-many ... *)
   
        val _      = (writeln("\n Computing the Convertables ... ");
                      writeln(  " ============================== \n"));

        val convertables = flat(map (convert root_thy subst) liftables)

        val defs         = lift0_def::lift1_def::lift2_def::
                              (map (fn (_,(_,t),_) => t) core_E)

        val _      = (writeln("\n Proving the Convertables ... ");
                      writeln(  " ============================ \n"));

        val errorfn_save = (!error_fn); 
        val _ = (error_fn:= (fn _ => ()))  (* proof engine: shut up *)                  
        val proven_lifts = Library.mapfilter (eq_lifter_prover root_thy thms defs) 
                                             (convertables)
        val _ = (error_fn:= errorfn_save)  (* proof engine: speak up again*)                  
        

        val _      = (writeln("\n Done ... ");
                      writeln(  " ======== \n"));

    in  (liftables,convertables,proven_lifts) end;


fun thy_morpher root_thy =
    let val (E,holthys)
                   = List.partition(fn x => subthy(theory "Lifting",x))
                            (root_thy::(ancestors_of root_thy));
        val _      = (writeln("\n Theory-Morpher from HOL to OCL: ");
                      writeln(  " =============================== \n"));

        val core_E = flat (map core E);
        val _      = if not(#trace_morpher(!state)) then () 
                     else let val H = Sign.string_of_typ thy
                              val L = map (fn ((Const(s,t),Const(s',t')),_,_) => 
                                               (s^"::"^(H t),s'^"::"^(H t'))) core_E
                              val _ = (writeln("\n  The core of the morphism: ");
                                       writeln("  >>>>>>>>>>>>>>>>>>>>>>>>"));
                              val _ = app (fn (s,t) => writeln("\n      "^s^"\n |==> "^t)) L
                              val _ = writeln("\n  <<<<<<<<<<<<<<<<<<<<<<<< \n");
                          in  () end

        val thms   = flat (map PureThy.thms_of holthys);
        val (liftables,convertables,proven_lifts)      
                   = lift root_thy core_E thms;

        fun add_thm ((name,thm),thy) = 
                     let val clean_dot = String.map(fn x=>if x= #"." then #"_" else x)
                     in  fst(PureThy.add_thms [((clean_dot("LIFT_"^name),thm),[])] thy) end

    in  state :=  {trace_morpher = #trace_morpher(!state),
                   thms          = thms,
                   liftables     = liftables,
                   convertables  = convertables,
                   proven_lifts  = proven_lifts};
        foldl add_thm root_thy proven_lifts
    end;


fun global_statistics () = 
    let val OCL_thms_no = length(PureThy.all_thms_of(the_context()))-
                          length(PureThy.all_thms_of(theory"Complex_Main")) 
        val {trace_morpher,thms,
             liftables,convertables,
             proven_lifts} = !state
    in  writeln ("Global Statistics: Number of OCL theorems: "^(Int.toString OCL_thms_no));

        writeln("\n theorem lifting:");
        writeln(" ================== \n");
        writeln(" HOL Theorems : "^Int.toString(length thms));
        writeln(" Liftables    : "^Int.toString(length liftables));
        writeln(" Convertables : "^Int.toString(length convertables));
        writeln(" Proven Lifts : "^Int.toString(length proven_lifts))
    end


    fun self_test () = 
        let val {trace_morpher, thms, 
             liftables, convertables, proven_lifts} = !state
        in  if length(liftables) < 550   then error"Thy_morpher Check: liftables"  else
            if length(convertables) < 736 then error"Thy_morpher Check: convertables"  else
            if length(proven_lifts) < 204 then error"Thy_morpher Check: proven_lifts"  else ()
        end;

*}

text{* Activation of the Theory-Morpher: *}
ML{*  val _ = thy_morpher (the_context());
      global_statistics ()
*}



(* Testing Zone 
ML{*

        val root_thy = the_context();

        val (E,holthys)
                   = List.partition(fn x => subthy(theory "Lifting",x))
                            (root_thy::(ancestors_of root_thy));
        val _      = (writeln("\n Theory-Morpher from HOL to OCL: ");
                      writeln(  " =============================== \n"));

        val core_E = flat (map core E);

*}
ML{*
fun show_signature_morphism core_E =
    let fun kk ((Const(s,_),Const(s',_)),(ss,_),_) = 
                  writeln(s^" --> "^s'^" "^ss)
           |kk ((Const(s,_), _),_,_) = writeln(s^" --> MURKS")
    in  app kk core_E end;

show_signature_morphism core_E;
*}

ML{* show_signature_morphism core_E;*}



ML{* (* Testing Zone *)

val root_thy = the_context();

        val (E,holthys)
                   = List.partition(fn x => subthy(theory "Lifting",x))
                            (root_thy::(ancestors_of root_thy));
        val core_E = flat (map core E);
        val thms   = flat (map PureThy.thms_of holthys);
*}

ML{* length(#proven_lifts(!state)) *}
ML{* thy_morpher (the_context()) *}


ML{*
(*  Testing Zone:
        val (liftables,convertables,proven_lifts)      
                   = lift root_thy core_E thms;
 *)

        val the_logical_connectives = ["Trueprop","op =", "Lifting.DEF",
	                               "All", "==>"]
	val core_E_M     = filter (fn ((_,Const(s,_)),_,_) => 
	                               forall(fn x => not(x = s))
				             (the_logical_connectives))
			          (core_E)
        val the_ran_core_E_M'       = map dest_Const (ran core_E_M)
        fun is_liftable (name,thm)  = 
                           let val cc  = consts_of(term_of(cprop_of thm)) 
                               val cc1 = subtract cc the_logical_connectives
                               val cc2 = subtract' cc1 the_ran_core_E_M'
                           in  not(null(cc1)) andalso null(cc2) end
        fun inst thm t  = (let val h as (_::_) = (term_tvars(concl_of thm))
                               val tt = ctyp_of (theory_of_thm thm)
                               fun convT (x as (s,["HOL.type"])) = raise Match
                                  |convT (x as (s,_)) = (tt(TVar x), tt(t))
                           in  SOME(instantiate (map convT h,[]) thm) end
                           handle _ => NONE);

        fun instantiate_classes (name,thm) = 
                            (case Library.mapfilter 
                                       (inst thm) 
                                       [HOLogic.intT,HOLogic.realT] of
                               [] => [(name,thm)]
                             | S  => map (fn x => (name,x)) S);  

        val liftables    = flat(map instantiate_classes
                                    (filter is_liftable thms));

        fun regrp ((x,y as Const(s,t)),a,parms) = (s,((x,y),(a,parms)))
    
        val subst        = Symtab.make_multi(map regrp core_E_M)
                           (* subst is the inversion of the core mapping.
                              Since the core of the morphism is *NOT*
                              injective in general, the inversion may be one-to-many ... *)
   
        val convertables = flat(map (convert root_thy subst) liftables)


*}

*)

section {* \acs{isar} Setup *}
ML {*
structure P = OuterParse and K = OuterKeyword;

fun apply_thymorpherT name thy = thy_morpher thy

val apply_thymorpherP =
      OuterSyntax.command "apply_thymorpher" "apply theory morpher"
                        OuterKeyword.thy_script
                        ((Scan.optional P.name "") >>(Toplevel.theory o apply_thymorpherT));

val _ =  OuterSyntax.add_parsers [apply_thymorpherP];
*}


text {* attributes \verb|cp_fold| and \verb|cp_unfold| *}
ML {*

fun cpFold_add name (thy,thm) =  (thy,thm)    (* stub! *)

fun cpUnfold_add name (thy,thm) =  (thy,thm)  (* stub! *)


fun cpFold_name x = Attrib.syntax (Scan.lift Args.name >> cpFold_add) x;

fun cpUnfold_name x = Attrib.syntax (Scan.lift Args.name >> cpUnfold_add) x;

val ThyMorpherSetup = [Attrib.add_attributes [("cp_fold",
                          (cpFold_name,
                           K Attrib.undef_local_attribute),
                          "declare theorems for cp_fold")],
                       Attrib.add_attributes [("cp_unfold",
                          (cpUnfold_name,
                           K Attrib.undef_local_attribute),
                          "declare theorems for cp_unfold")]
                       ];
*}

setup "ThyMorpherSetup"
(*>*)

end
 
