(*****************************************************************************
 *          HOL-OCL - a shallow embedding of OCL in Isabelle/HOL              
 *                                                                            
 * OCL_Bag.thy -
 * Copyright (C) 2001,2002,2003,2004 Achim D. Brucker <brucker@inf.ethz.ch>   
 *                                   Burkhart Wolff   <bwolff@inf.ethz.ch>    
 *                                                                            
 * This file is part of HOL-OCL.                                              
 *                                                                            
 * HOL-OCL is free software; you can redistribute it and/or modify it under   
 * the terms of the GNU General Public License as published by the Free       
 * Software Foundation; either version 2 of the License, or (at your option)  
 * any later version.                                                         
 *                                                                            
 * HOL-OCL is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS 
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more 
 * details.                                                              
 *                                                                            
 * You should have received a copy of the GNU General Public License along    
 * with this program; if not, write to the Free Software Foundation, Inc.,    
 * 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.                  
 ******************************************************************************)

header {* \HOL Collection Extensions *}

theory HOL_Collections_ext
imports 
  "Library/Permutation"
  "Library/Multiset" 
begin


text{* Redefinition of the overwritten theorems in the Permutation theory *}

lemmas trans = HOL.trans
lemmas swap = HOL.swap


section{* Extending the HOL list theory *}

subsection{* Further case distinctions and induction schemes*}

text{* These two theorems which have an explicit singleton case
  are very handy when proving theorems about $remstutter$ which has
  three pattern matching cases.
*}

lemma list_single_cases:
  assumes emptyC: "xs = [] \<Longrightarrow> P xs"
  and    singleC: "\<And> a. xs = [a] \<Longrightarrow> P xs"
  and      stepC: "\<And> a b ys. xs = a#b#ys \<Longrightarrow> P xs"
  shows "P xs"
  apply(cases xs, erule emptyC)
  apply(case_tac list, rule singleC, simp)
  apply(rule_tac a=a and b=aa and ys=lista in stepC, simp)
  done

lemma list_single_induct:
  "\<lbrakk> P []; \<And> a.  P [a]; \<And> a b ys. \<lbrakk> P (b#ys); P ys \<rbrakk> \<Longrightarrow> P (a#b#ys) \<rbrakk> \<Longrightarrow> P xs"
  by(rule length_induct, rule list_single_cases, simp_all)


lemma list_rev_induct2:
  "\<lbrakk>length xs = length ys; P [] [];
    \<And> x xs y ys. \<lbrakk> length xs = length ys; P xs ys \<rbrakk> \<Longrightarrow> P (xs @ [x]) (ys @ [y])\<rbrakk>
   \<Longrightarrow> P xs ys"
  apply(simplesubst rev_rev_ident[symmetric])
  apply(subst rev_rev_ident[symmetric])
  apply(rule_tac xs="rev xs" and ys="rev ys" in list_induct2)
  apply(simp_all)
  done


lemma unzip_zip_id:
  "\<lbrakk> length xs = length ys; unzip = (\<lambda> x. (map fst x, map snd x)); curried = (\<lambda> x. P (fst x) (snd x))  \<rbrakk> \<Longrightarrow>
   P xs ys = curried (unzip (zip xs ys))"
  by(simp)


lemma length_induct2:
  "\<lbrakk> length as = length bs;
     (\<forall> xs ys. length xs = length ys \<and>
         (\<forall> ws vs. 
            ((length ws < length xs) \<and> (length ws = length vs)) \<longrightarrow> P ws vs) 
         \<longrightarrow> P xs ys) \<rbrakk>
   \<Longrightarrow> P as bs"
  apply(rule_tac t="P as bs" in subst, rule sym[OF unzip_zip_id])
  apply(simp) prefer 3
  apply(rule_tac xs="(zip as bs)" in length_induct)
  prefer 3 prefer 3
  apply(rule refl)+
  apply(erule_tac x="(fst ((map fst xs), (map snd xs)))" in allE)
  apply(rotate_tac -1)
  apply(erule_tac x="(snd ((map fst xs), (map snd xs)))" in allE)
  apply(erule mp)
  apply(auto)
  apply(erule_tac x="zip ws vs" in allE)
  apply(simp)
  done


subsection{* Theorems concerning already existing functions on lists *}


subsubsection{* last *}

lemma last_rev: "xs \<noteq> [] \<Longrightarrow> last (rev xs) = hd xs"
  by(cases xs, auto)


subsubsection{* remdups *}

lemma remdups_append [simp]:
  "set xs \<inter> set ys = {} \<Longrightarrow> remdups (xs@ys) = (remdups xs) @ (remdups ys)"
  by(induct xs, auto)


subsubsection{* take and drop *}

lemma notin_set_take:
  "x \<notin> set xs \<Longrightarrow> x \<notin> set(take n xs)"
  apply(rule contra_subsetD)
  apply(rule set_take_subset, assumption)
  done

lemma notin_set_drop:
  "x \<notin> set xs \<Longrightarrow> x \<notin> set(List.drop n xs)"
  apply(rule contra_subsetD)
  apply(rule set_drop_subset, assumption)
  done


subsection{* A very often occurring filter expression: $filter\ (\lambda\ x.\ x = a)$ *}

text{* This filter expressions occurs very frequently when reasoning
  over $remstutter$. Because it actually represents just a list of
  equal values, we will refer to it as a \emph{stutter}. The prefix
  \emph{filter\_simple\_eq} is due to the simple structure of the
  expressions. Feel free if to suggest another name, if you know one.
*}

lemma filter_simple_eq_exhaustion:
  "(List.filter (\<lambda> x. x = a) xs  = []) \<or> 
   (\<exists> ys. List.filter (\<lambda> x. x = a) xs = a # List.filter (\<lambda> y. y = a) ys
                            \<and> length ys < length xs)"
  apply(case_tac "List.filter (\<lambda> x. x = a) xs", simp_all)
  apply(drule filter_eq_ConsD, auto)
  done

lemmas filter_simple_eq_cases = disjE[OF filter_simple_eq_exhaustion, standard]

lemma filter_simple_eq_rev_exhaustion:
  "(List.filter (\<lambda> x. x = a) xs  = []) \<or> 
   (\<exists> ys. List.filter (\<lambda> x. x = a) xs = (List.filter (\<lambda> y. y = a) ys) @ [a]
                            \<and> length ys < length xs)"
  apply(rule_tac a=a and xs="rev xs" in filter_simple_eq_cases)
  apply(simp add: filter_empty_conv)
  apply(rule disjI2, clarify)
  apply(drule_tac f=rev in arg_cong)
  apply(auto simp: rev_filter)
  done

lemmas filter_simple_eq_rev_cases = disjE[OF filter_simple_eq_rev_exhaustion, standard]

lemma list_stutter_exhaustion:
  "xs = [] \<or> (\<exists> ys zs a list. xs=a#list \<and> 
                xs = List.filter (\<lambda> y. y = a) ys @ zs \<and> (hd zs \<noteq> a \<or> zs = [])
                \<and> a \<in> set ys \<and> length zs < length xs)"
  apply(rule_tac xs=xs in list_single_induct, simp_all)
  apply(rule_tac x="[a]" in exI, simp)
  apply(case_tac "b \<noteq> a")
  apply(rule_tac x="[a]" in exI, simp_all)
  apply(clarify)
  apply(rule_tac x="a#ysa" in exI, simp)
  done

lemmas list_stutter_cases = disjE[OF list_stutter_exhaustion, standard]


lemma filter_simple_eq_empty [simp]: "a \<notin> set xs \<Longrightarrow> List.filter (\<lambda> x. x = a) xs = []"
  by(auto simp: filter_empty_conv)

lemma filter_simple_eq_cons: 
  "a \<in> set xs \<Longrightarrow> \<exists> ys. List.filter (\<lambda> x. x = a) xs = a # List.filter (\<lambda> y. y = a) ys 
                       \<and> length ys < length xs"
  apply(case_tac "List.filter (\<lambda> x. x = a) xs")
  apply(simp add: filter_empty_conv, blast)
  apply(drule filter_eq_ConsD, auto)
  done


lemma hd_filter_simple_eq [simp]: "a \<in> set xs \<Longrightarrow> hd (List.filter (\<lambda> x. x = a) xs) = a"
  apply(erule rev_mp)
  apply(rule_tac xs=xs in length_induct)
  apply(rule_tac xs=xs and a=a in filter_simple_eq_cases)
  apply(auto simp: filter_empty_conv)
  done

lemma last_filter_simple_eq [simp]: "a \<in> set xs \<Longrightarrow> last (List.filter (\<lambda> x. x = a) xs) = a"
  apply(subst rev_rev_ident[symmetric])
  apply(subst rev_filter)
  apply(auto simp: filter_empty_conv last_rev)
  done

lemma rev_filter_simple_eq [simp]: 
  "rev (List.filter (\<lambda> x. x = a) xs) = (List.filter (\<lambda> x. x = a) xs)"
  apply(rule_tac xs=xs in length_induct)
  apply(rule_tac a=a and xs=xs in filter_simple_eq_cases, simp)
  apply(clarify, simp only:, simp (no_asm_use))
  apply(rule_tac a=a and xs=ys in filter_simple_eq_rev_cases, simp)
  apply(clarify, simp)
  done


subsection{* remstutter *}

text{* remstutter removes all ``stutters'', i.e., consecutive
  sequences of equal elements. Thus two lists $xs$ and $ys$ are \emph{stutter equivalent} 
  if it holds that $remstutter\ xs = remstutter\ ys$.
*}

consts
  remstutter :: "'a list \<Rightarrow> 'a list"

primrec
  "remstutter [] = []"
  "remstutter (x#xs) = (case xs of   [] \<Rightarrow> [x]
                                 | y#ys \<Rightarrow> (if x=y then     remstutter xs 

                                                   else x # remstutter xs))"

lemma remstutter_simps:
  "remstutter [] = []"
  "remstutter [a] = [a]"
  "remstutter (a#b#xs) = (if a=b then remstutter (b#xs) else a # remstutter (b#xs))"
  by(simp_all)

text{*
  To provide a convenient (i.e., readable) simplification during 
  proofs about remstutter we redefine the simplification rules.
*}

declare remstutter.simps [simp del]
declare remstutter_simps [simp]


lemma remstutter_cons1: "\<lbrakk> x = hd xs; xs \<noteq> [] \<rbrakk> \<Longrightarrow> remstutter (x#xs) = remstutter xs"
  by(case_tac xs, simp_all)

lemma remstutter_cons2: "\<lbrakk> x \<noteq> hd xs; xs \<noteq> [] \<rbrakk> \<Longrightarrow> remstutter (x#xs) = x # remstutter xs"
  by(case_tac xs, simp_all)

lemma set_remstutter [simp]: "set (remstutter xs) = set xs"
  by(rule_tac xs=xs in list_single_induct, auto)

lemma length_remstutter_leq [iff]: "length (remstutter xs) <= length xs"
  by(rule_tac xs=xs in list_single_induct, auto)

lemma remstutter_eq_nil_iff [simp]: "(remstutter xs = []) = (xs = [])"
  by(induct xs, simp, case_tac xs, auto)

lemma remstutter_eq_nil_right_iff [simp]: "([] = remstutter xs) = (xs = [])"
  by(induct xs, simp, case_tac xs, auto)

lemma remstutter_if_distinct: 
  "distinct xs \<Longrightarrow> remstutter xs = xs"
  by(erule rev_mp, rule_tac xs=xs in list_single_induct, auto)

lemma length_remstutter_eq [iff]:
  "(length (remstutter xs) = length xs) = (remstutter xs = xs)"
  oops

lemma distinct_remstutter [simp]:
  "distinct xs \<Longrightarrow> distinct (remstutter xs)"
  by(erule rev_mp, rule_tac xs=xs in list_single_induct, auto)

lemma remstutter_eq_remdups: 
  "distinct (remstutter xs) \<Longrightarrow> remstutter xs = remdups xs"
  by(erule rev_mp, rule_tac xs=xs in list_single_induct, auto)

lemma remstutter_remdups: "remstutter (remdups xs) = remdups xs"
  by(simp add: remstutter_if_distinct)

lemma remdups_remstutter [simp]: "remdups (remstutter xs) = remdups xs"
  by(rule_tac xs=xs in list_single_induct, simp_all)

lemma remstutter_append:
  "last xs \<noteq> hd ys \<Longrightarrow> remstutter (xs @ ys) = (remstutter xs) @ (remstutter ys)"
  apply(erule rev_mp)
  apply(rule_tac xs=ys in list_single_induct, simp_all)
  apply(rule_tac xs=xs in list_single_induct, simp_all)+
  done

lemma remstutter_stutter [simp]:
  "a \<in> set xs \<Longrightarrow> remstutter (List.filter (\<lambda> x. x = a) xs) = [a]"
  apply(erule rev_mp)
  apply(rule_tac xs=xs in length_induct, clarify)
  apply(rule_tac xs=xs and a=a in filter_simple_eq_cases)
  apply(simp add: filter_empty_conv, blast)
  apply(clarify)
  apply(erule_tac x=ys in allE)
  apply(case_tac "a \<in> set ys", simp)
  apply(rule_tac xs=ys and a=a in filter_simple_eq_cases)  
  apply(simp)
  apply(clarify, simp)
  apply(simp add: filter_simple_eq_empty)
  done

lemma remstutter_stutter_prepend:
  "\<lbrakk> a \<in> set ys; hd zs \<noteq> a \<or> zs = [] \<rbrakk> \<Longrightarrow>
   remstutter (List.filter (\<lambda> y. y = a) ys @ zs) = a # remstutter zs"
  by(auto simp: remstutter_append neq_commute)


text{* There hold quite a few additional lemmas for sequences that 
  are distinct after the removal of all ``stutters''. *}

lemma distinct_remstutter_ConsD:
  "distinct (remstutter (x # xs)) \<Longrightarrow> distinct (remstutter xs)"
  by(erule contrapos_pp, rule_tac xs=xs in list_single_cases, simp_all)

lemma distinct_remstutter_appendD1:
  "distinct (remstutter (xs@ys)) \<Longrightarrow> distinct (remstutter xs)"
  by(erule rev_mp, rule_tac xs=xs in list_single_induct, auto)

lemma distinct_remstutter_appendD2:
  "distinct (remstutter (xs@ys)) \<Longrightarrow> distinct (remstutter ys)"
  by(induct xs, auto dest: distinct_remstutter_ConsD)

lemma distinct_remstutter_hd_cases:
  " \<lbrakk> distinct(remstutter xs) \<rbrakk> \<Longrightarrow>
    xs = [] \<or> (xs \<noteq> [] \<and> (\<exists> ys. xs = List.filter (\<lambda> x. x = hd xs) xs @ ys
                        \<and> hd xs \<notin> set ys
                        \<and> distinct(remstutter ys) \<and> length ys < length xs))"
  apply(erule rev_mp)
  apply(rule_tac xs=xs in list_single_induct)
  apply(simp_all add: neq_commute)
  apply(clarify)
  apply(rule_tac x="ysa" in exI, simp)
  done

lemma distinct_remstutter_decomp:
  "\<And> a. distinct (remstutter vs) \<Longrightarrow>
   \<exists> xs ys zs. vs = xs @ ys @ zs \<and> a \<notin> set xs \<and> a \<notin> set zs 
               \<and> ys = List.filter (\<lambda> v. v = a) vs"
  apply(case_tac "a \<notin> set vs")
  apply(rule_tac x=vs in exI)
  apply(simp_all add: filter_simple_eq_empty)
  apply(erule rev_mp)+
  apply(rule_tac xs=vs in length_induct)
  apply(clarify)
  apply(frule distinct_remstutter_hd_cases)
  apply(case_tac xs, simp, simp)
  apply(rule conjI, clarify)
  apply(rule_tac x="[]" in exI)
  apply(rule_tac x="ys" in exI, simp)
  apply(clarify)
  apply(frule distinct_remstutter_ConsD)
  apply(erule_tac x="list" in allE, simp add: neq_commute)
  apply(clarify)
  apply(rule_tac x="aa#xs" in exI)
  apply(rule_tac x="zs" in exI)
  apply(simp)
  done

lemma rev_remstutter: "rev (remstutter xs) = remstutter (rev xs)"
  apply(rule_tac xs=xs in  length_induct)
  apply(rule_tac xs=xs in list_stutter_cases, simp)
  apply(clarify, simp)
  apply(case_tac "zs=[]", simp)
  apply(erule_tac x="zs" in allE)
  apply(simp add: neq_commute last_rev remstutter_append)
  done

lemma rev_remdups: "distinct (remstutter xs) \<Longrightarrow>
  rev (remdups xs) = remdups (rev xs)"
  apply(subgoal_tac "distinct (remstutter (rev xs))")
  apply(simp add: remstutter_eq_remdups[symmetric] rev_remstutter)
  apply(simp add: rev_remstutter[symmetric])
  done


section{* Extending the HOL Multiset theory *}


subsection{* further induction schemes *}

lemma multiset_size_induct:
  "\<lbrakk> \<And> M. \<forall> N. size N < size M \<longrightarrow> P N \<Longrightarrow> P M \<rbrakk> \<Longrightarrow> P (M::'a multiset)"
  by(rule_tac f="size::('a multiset \<Rightarrow> nat)" in measure_induct) iprover


subsection{* empty multiset *}

lemma set_of_eq_empty_conv1: "(set_of M = {}) = (M = {#})"
  by(auto)

lemma set_of_eq_empty_conv2 [simp]: "({} = set_of M) = (M = {#})"
  by(rule iffI, drule sym, auto)


subsection{* union *}

lemma diff_union_inverse1: "0 < count M x \<Longrightarrow> M - {#x#} + {#x#} = M"
  by(simp add: multiset_eq_conv_count_eq)


subsection{* MCollect *}

lemma MCollect_empty [simp]:
  "{# x:{#}. P x #} = {#}"
  apply(simp add: MCollect_def Mempty_def )
  apply(subst Abs_multiset_inverse, simp_all)
  done

lemma MCollect_single_false [simp]:
  "\<not> P a \<Longrightarrow> {# x: (M + {#a#}). P x #} = {# x: M. P x #}"
  apply(simp add: MCollect_def union_def single_def
                  Abs_multiset_inject Rep_multiset 
                  MCollect_preserves_multiset)
  apply(rule ext)
  apply(simp add: Abs_multiset_inverse Rep_multiset MCollect_preserves_multiset)
  done

lemma MCollect_single_true [simp]:
  "P a \<Longrightarrow> {# x: (M + {#a#}). P x #} = {# x: M. P x #} + {#a#}"
  apply(simp add: MCollect_def union_def single_def
                  Abs_multiset_inject Rep_multiset 
                  MCollect_preserves_multiset)
  apply(rule ext)
  apply(simp add: Abs_multiset_inverse Rep_multiset MCollect_preserves_multiset)
  done


subsection{* multiset_of *}

lemma multiset_of_filter [simp]: 
  "multiset_of (List.filter P xs) = {# x : (multiset_of xs). P x #}"
  by(induct xs, auto)



section{* Connecting the theories: add the missing conversions*}

subsection{* From finite sets to lists *}

text{* A \emph{finite} set is converted to a list by taking just \emph{some} 
  list that is distinct and contains the same elements like the set *}

constdefs 
  list_of_set ::   "'a set \<Rightarrow> 'a list"
  list_of_set_def: "list_of_set X \<equiv> SOME ys. distinct ys \<and> set ys = X"


lemma list_of_setI:
  "\<lbrakk> finite X; \<And> ys. \<lbrakk> distinct ys; set ys = X \<rbrakk> \<Longrightarrow> P ys \<rbrakk> \<Longrightarrow>
   P (list_of_set X)"
  apply(unfold list_of_set_def)
  apply(rule someI2_ex)
  apply(rule finite_induct, auto)
  apply(rule_tac x="x#a" in exI, simp)
  done

lemma set_list_of_set [simp]: 
  "finite X \<Longrightarrow> set (list_of_set X) = X"
  by(rule list_of_setI, simp_all)

lemma distinct_list_of_set [simp]: 
  "finite X \<Longrightarrow> distinct (list_of_set X)"
  by(rule list_of_setI, simp_all)

lemma list_of_set_empty [simp]:
  "list_of_set {} = []"
  by(rule list_of_setI, auto)

lemma list_of_set_eq_empty_iff [simp]:
  "finite X \<Longrightarrow> (list_of_set X = []) = (X = {})"
  by(erule list_of_setI, auto)

lemma list_of_set_single [simp]: "list_of_set {a} = [a]"
  apply(rule list_of_setI, simp)
  apply(case_tac ys, simp)
  apply(case_tac list, auto)
  done

lemma list_of_set_eq_single_iff [simp]:
  "finite X \<Longrightarrow> (list_of_set X = [a]) = (X = {a})"
  apply(erule list_of_setI)
  apply(rule iffI, simp)
  apply(case_tac ys, simp)
  apply(case_tac list, auto)
  done


subsection{* From sets to multisets *}

constdefs 
  multiset_of_set ::   "'a set \<Rightarrow> 'a multiset"
  multiset_of_set_def: "multiset_of_set X \<equiv> Abs_multiset (\<lambda>a. if (a \<in> X) then 1 else 0)"

lemma multiset_of_set_in_multiset [simp]:
  "finite X \<Longrightarrow> (\<lambda>a. if (a \<in> X) then 1 else 0) \<in> multiset"
  by(simp add: multiset_def)

lemma count_multiset_of_set:
  "finite X \<Longrightarrow> (count (multiset_of_set X) a) = (if (a \<in> X) then 1 else 0)"
  by(simp add: multiset_of_set_def count_def Abs_multiset_inverse)

lemma in_multiset_of_set [simp]:
  "finite X \<Longrightarrow> (0 < count (multiset_of_set X) a) = (a \<in> X)"
  by(simp add: count_multiset_of_set)

lemma notin_multiset_of_set [simp]:
  "finite X \<Longrightarrow> (count (multiset_of_set X) a = 0) = (a \<notin> X)"
  by(simp add: count_multiset_of_set)

lemma set_of_multiset_of_set [simp]:
  "finite X \<Longrightarrow> set_of (multiset_of_set X) = X"
  by(simp add: set_of_def)

lemma multiset_of_set_empty [simp]: "multiset_of_set {} = {#}"
  by(simp add: multiset_eq_conv_count_eq)

lemma multiset_of_set_insert [simp]: 
  "\<lbrakk> finite X; a \<notin> X \<rbrakk> \<Longrightarrow> multiset_of_set (insert a X) = (multiset_of_set X) + {#a#}"
  by(simp add: multiset_eq_conv_count_eq count_multiset_of_set neq_commute)

lemma multiset_of_set_union:
  "\<lbrakk> finite A; finite B \<rbrakk> \<Longrightarrow> 
   multiset_of_set (A \<union> B) = (multiset_of_set A) + (multiset_of_set (B - A))"
  by(simp add: multiset_eq_conv_count_eq count_multiset_of_set neq_commute)

lemma multiset_of_set_set_distinct [simp]:
  "distinct xs \<Longrightarrow> multiset_of_set (set xs) = multiset_of xs"
  by(induct xs, simp_all)


subsection{* From multisets to lists *}

text{* To convert a multiset to \emph{some} list actually any list
  that contains exactly the same number of each element in the 
  multiset would suit. To get a stronger structure with respect to
  the conversion of finite sets to lists we chose to order all
  elements of the multiset according to the ``order'' given by
  the conversion of the set of all distinct elements of the 
  multiset to a list. Thus resulting in what we call a ``doublette_list''.
*}

constdefs 
  list_of_multiset     ::   "'a multiset \<Rightarrow> 'a list"
  list_of_multiset_def : 
  "list_of_multiset M  \<equiv> SOME xs. multiset_of xs = M \<and> remstutter xs = (list_of_set (set_of M))"

lemma multiset_as_doublette_list:
 "\<lbrakk> distinct ys; set ys = set_of M\<rbrakk>
  \<Longrightarrow> \<exists> xs. multiset_of xs = M \<and> remstutter xs = ys"
  apply(erule rev_mp)+
  apply(rule_tac P="\<lambda> M::'a multiset. ?P M" in spec)
  apply(induct_tac ys, simp, clarify)
  apply(erule_tac x="{# y:x. y \<noteq> a #}" in allE, simp)
  apply(subgoal_tac "set list = set_of x \<inter> {y. (y \<noteq> a)}")
  apply(subgoal_tac "\<exists> ys. x = multiset_of ys")
  apply(simp, clarify)
  apply(rule_tac x="(List.filter (\<lambda> y. y = a) ys)@xs" in exI)
  apply(simp add: union_ac)
  apply(subst remstutter_stutter_prepend) defer 1
  apply(case_tac xs) defer 1 defer 1
  apply(simp add: trans[OF multiset_partition[symmetric]])
  apply(rule surjD[OF surj_multiset_of])
  apply(auto)
  done

lemma list_of_multisetI:
  assumes doublette_repr: "\<And> xs. \<lbrakk> M = multiset_of xs; distinct(remstutter xs) \<rbrakk> \<Longrightarrow> P xs"
  shows "P (list_of_multiset M)"
  apply(unfold list_of_multiset_def)
  apply(rule someI2_ex, rule list_of_setI, simp)
  apply(rule multiset_as_doublette_list, simp_all)
  apply(rule_tac X="set_of M" in list_of_setI, simp)
  apply(rule doublette_repr, simp_all)
  done

lemma set_list_of_multiset [simp]:
  "set (list_of_multiset M) = set_of M"
  by(rule list_of_multisetI, simp)

lemma multiset_list_of_multiset [simp]:
  "multiset_of (list_of_multiset M) =  M"
  by(rule list_of_multisetI, simp)

lemma list_of_multiset_multiset_of_set [simp]:
  "finite X \<Longrightarrow> list_of_multiset (multiset_of_set X) = list_of_set X"
  apply(simp add: list_of_multiset_def)
  apply(rule someI2_ex[OF multiset_as_doublette_list])
  apply(simp_all)
  apply(subgoal_tac "distinct x")
  apply(simp add: remstutter_if_distinct)
  apply(simp add: distinct_count_atmost_1 count_multiset_of_set 
                  multiset_eq_conv_count_eq mem_set_multiset_eq)
  done

lemma list_of_multiset_empty [simp]:
  "list_of_multiset {#} = []"
  by(rule list_of_multisetI, simp)

lemma list_of_multiset_eq_empty_iff [simp]:
  "(list_of_multiset M = []) = (M = {#})"
  by(rule list_of_multisetI, simp)

lemma list_of_multiset_single [simp]: "list_of_multiset {#a#} = [a]"
  apply(rule list_of_multisetI, case_tac xs)
  by(simp_all add: single_is_union)

lemma list_of_multiset_eq_single_iff [simp]:
  "(list_of_multiset M = [a]) = (M = {#a#})"
  apply(rule list_of_multisetI)
  apply(rule iffI, simp)
  apply(case_tac xs, simp)
  apply(simp add: single_is_union)
  done

lemma remdups_list_of_multiset [simp]:
  "remdups (list_of_multiset M) = list_of_set (set_of M)"
  apply(unfold list_of_multiset_def, rule someI2_ex, rule list_of_setI)
  by(simp_all add: multiset_as_doublette_list remstutter_eq_remdups[symmetric])

lemma distinct_remstutter_list_of_multiset [simp]:
  "distinct (remstutter (list_of_multiset M))"
  apply(unfold list_of_multiset_def, rule someI2_ex, rule list_of_setI)
  by(simp_all add: multiset_as_doublette_list)




subsection{* fold on multisets *}

constdefs 
  fold_multiset     ::   "('b \<Rightarrow> 'a \<Rightarrow> 'b) \<Rightarrow> 'b \<Rightarrow> 'a multiset \<Rightarrow> 'b"
  fold_multiset_def : 
  "fold_multiset f e M \<equiv> foldl f e (list_of_multiset M)"


lemma fold_multiset_empty [simp]: "fold_multiset f e {#} = e"
  by(simp add: fold_multiset_def)

lemma fold_multiset_single [simp]: "fold_multiset f e {#a#} = f e a"
  by(simp add: fold_multiset_def)

text{* the two core theorems that enable us to use order independence *}
lemma foldr_left_order_independent:
  "\<lbrakk> multiset_of xa = multiset_of xs; \<forall> a b c. (f a (f b c)) = (f b (f a c)) \<rbrakk> 
   \<Longrightarrow> foldr f xa e = foldr f xs e"
  apply(subst (asm) multiset_of_eq_perm)
  apply(frule rev_mp) prefer 2
  apply(assumption)
  apply(rule_tac xb="xa" and xa="xs" in perm.induct)
  apply(simp_all)
  done

lemma foldl_right_order_independent:
  "\<lbrakk> multiset_of xa = multiset_of xs; \<forall> a b c. (f (f a b) c) = (f (f a c) b) \<rbrakk> 
  \<Longrightarrow> foldl f e xa = foldl f e xs"
  apply(subst foldl_foldr)+
  apply(rule foldr_left_order_independent)
  apply(subst (asm) multiset_of_eq_perm)
  apply(subst multiset_of_eq_perm)
  apply(rule Permutation.trans, rule perm_rev)
  apply(rule perm_sym, rule Permutation.trans, rule perm_rev)
  apply(rule perm_sym, simp_all)
  done

text{* This enables us to unfold a union *}
lemma fold_multiset_union [simp]:
  assumes ordIndep: "\<forall> a b c. (f (f a b) c) = (f (f a c) b)"
  shows "fold_multiset f e (M + N) = (fold_multiset f (fold_multiset f e M) N)"
  apply(unfold fold_multiset_def)
  apply(rule list_of_multisetI, rule list_of_multisetI)
  apply(rule list_of_multisetI)
  apply(subgoal_tac "multiset_of (xsa @ xs) = multiset_of xsb")
  apply(drule_tac f=f and e=e in foldl_right_order_independent)
  apply(rule ordIndep, simp_all add: union_commute)
  done

lemma fold_multiset_union_rev:
  assumes ordIndep: "\<forall> a b c. (f (f a b) c) = (f (f a c) b)"
  shows "fold_multiset f e (M + N) = (fold_multiset f (fold_multiset f e N) M)"
  by(subst union_commute, simp add: ordIndep)


subsection{* fold on sets *}

constdefs 
  fold_set     ::   "('b \<Rightarrow> 'a \<Rightarrow> 'b) \<Rightarrow> 'b \<Rightarrow> 'a set \<Rightarrow> 'b"
  fold_set_def : 
  "fold_set f e X \<equiv> foldl f e (list_of_set X)"

lemma fold_set_empty [simp]: "fold_set f e {} = e"
  by(simp add: fold_set_def)

lemma fold_set_fold_multiset_conv:
  "finite X \<Longrightarrow> 
   fold_set f e X = fold_multiset f e (multiset_of_set X)"
  by(simp add: fold_set_def fold_multiset_def)

lemma fold_set_single [simp]: "fold_set f e {a} = f e a"
  by(simp add: fold_set_def)

lemma fold_set_insert [simp]:
  assumes finite: "finite X"
  and   not_Cont: "a \<notin> X"
  and   ordIndep: "\<forall> a b c. (f (f a b) c) = (f (f a c) b)"
  shows "fold_set f e (insert a X) = f (fold_set f e X) a"
  by(simp add: finite ordIndep not_Cont fold_set_fold_multiset_conv)

lemma (in ACf) fold_set_eq_Finite_Set_fold:
  assumes finite: "finite X"
  and   h_by_f_g: "h = (\<lambda> x y. f x (g y))"
  shows "fold_set h e X = fold f g e X"
  apply(rule finite_induct[OF finite])
  apply(simp_all add: fold_set_insert h_by_f_g AC)
  done

lemma fold_set_union [simp]:
  assumes ordIndep: "\<forall> a b c. (f (f a b) c) = (f (f a c) b)"
  and      finiteX: "finite X"
  and      finiteY: "finite Y"
  shows "fold_set f e (X \<union> Y) = (fold_set f (fold_set f e X) (Y-X))"
  by(simp add: fold_set_fold_multiset_conv ordIndep finiteX finiteY
               multiset_of_set_union)

lemma fold_set_union_rev:
  assumes ordIndep: "\<forall> a b c. (f (f a b) c) = (f (f a c) b)"
  and      finiteX: "finite X"
  and      finiteY: "finite Y"
  shows "fold_set f e (X \<union> Y) = (fold_set f (fold_set f e (Y-X)) X)"
  by(simp add: fold_set_fold_multiset_conv multiset_of_set_union 
               ordIndep finiteX finiteY fold_multiset_union_rev
          del: fold_set_union fold_multiset_union)



end